# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.materials import Concrete, ConcreteSteel
from pycivil.EXAUtils.markdownReportMakers import ConcreteMdFB, SteelConcreteMdFB
from pycivil.EXAUtils.report import (
    EnumFBSection,
    MarkdownReporter,
    ReportTemplateEnum,
    getMarkdownTemplatesPath,
)


def test_report_markdown_concrete_0001(tmp_path: Path):
    """Test concrete material markdown fragment generation."""
    code = Code("NTC2018")

    concrete = Concrete()
    concrete.setByCode(codeObj=code, catstr="C25/30")

    markdownTemplatePath = getMarkdownTemplatesPath()
    fragmentBuilder = ConcreteMdFB(markdownTemplatePath, concrete)

    fragmentBuilder.setFragmentOptions({"glossary": False})

    f = fragmentBuilder.buildFragment()

    reporter = MarkdownReporter(markdownTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.MD_ENG_CAL, fragments=[f], glossary=False
    )
    reporter.compileDocument(path=str(tmp_path), output_format="md")
    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.md").exists()

    # Verify content has concrete data
    content = tmp_path.joinpath("report.md").read_text(encoding="utf-8")
    assert "C25/30" in content or "25.0" in content


def test_report_markdown_concrete_0002(tmp_path: Path):
    """Test concrete material with HTML output."""
    code = Code("NTC2018")

    concrete = Concrete()
    concrete.setByCode(codeObj=code, catstr="C35/45")

    markdownTemplatePath = getMarkdownTemplatesPath()
    fragmentBuilder = ConcreteMdFB(markdownTemplatePath, concrete)

    f = fragmentBuilder.buildFragment()

    reporter = MarkdownReporter(markdownTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.MD_ENG_CAL, fragments=[f]
    )
    reporter.compileDocument(path=str(tmp_path), output_format="html")
    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.md").exists()
    assert tmp_path.joinpath("report.html").exists()


def test_report_markdown_steel_0001(tmp_path: Path):
    """Test steel material markdown fragment generation."""
    code = Code("NTC2018")

    steel = ConcreteSteel()
    steel.setByCode(code, "B450A")

    markdownTemplatePath = getMarkdownTemplatesPath()

    fragmentBuilder = SteelConcreteMdFB(markdownTemplatePath, steel)

    fragmentBuilder.setFragmentOptions(
        {"section_title": "Acciaio B450A", "section_level": EnumFBSection.SEC_SECTION}
    )
    f1 = fragmentBuilder.buildFragment()

    steel.setByCode(code, "B450C")
    fragmentBuilder = SteelConcreteMdFB(markdownTemplatePath, steel)

    fragmentBuilder.setFragmentOptions(
        {"section_title": "Acciaio B450C", "section_level": EnumFBSection.SEC_SECTION}
    )

    # Here unsetted by code, the builder strip code used and kind of sensitivity
    steel.set_gammas(1.0)
    f2 = fragmentBuilder.buildFragment()

    reporter = MarkdownReporter(markdownTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.MD_ENG_CAL, fragments=[f1, f2]
    )
    reporter.compileDocument(path=str(tmp_path), output_format="md")

    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.md").exists()

    # Verify content has steel data
    content = tmp_path.joinpath("report.md").read_text(encoding="utf-8")
    assert "B450" in content or "450" in content


def test_report_markdown_combined_materials(tmp_path: Path):
    """Test combined concrete and steel materials in one report."""
    code = Code("NTC2018")

    # Concrete
    concrete = Concrete()
    concrete.setByCode(codeObj=code, catstr="C25/30")
    markdownTemplatePath = getMarkdownTemplatesPath()
    concreteFB = ConcreteMdFB(markdownTemplatePath, concrete)
    concreteFB.setFragmentOptions({"section_title": "Calcestruzzo C25/30"})
    f1 = concreteFB.buildFragment()

    # Steel
    steel = ConcreteSteel()
    steel.setByCode(code, "B450C")
    steelFB = SteelConcreteMdFB(markdownTemplatePath, steel)
    steelFB.setFragmentOptions({"section_title": "Acciaio B450C"})
    f2 = steelFB.buildFragment()

    reporter = MarkdownReporter(markdownTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.MD_ENG_CAL, fragments=[f1, f2]
    )
    reporter.compileDocument(path=str(tmp_path), output_format="md")

    assert tmp_path.joinpath("report.md").exists()
    content = tmp_path.joinpath("report.md").read_text(encoding="utf-8")
    assert "C25/30" in content or "25.0" in content
    assert "B450" in content or "450" in content



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

"""Helper functions for Pydantic model operations."""

from typing import Any, TypeVar, Type
from pydantic import BaseModel, ValidationError

T = TypeVar("T", bound=BaseModel)


def find_all_submodels(data: dict[str, Any] | list[Any], model_class: Type[T]) -> list[T]:
    """Find all occurrences of a submodel in nested JSON data.

    Recursively searches through nested dictionaries and lists to find
    all data structures that can be validated as the given Pydantic model.

    Args:
        data: The JSON data structure (dict or list) to search.
        model_class: The Pydantic model class to search for.

    Returns:
        A list of all matching model instances found in the data.

    Example:
        >>> class Inner(BaseModel):
        ...     x: int
        ...     y: str
        >>> data = {"items": [{"x": 1, "y": "a"}, {"x": 2, "y": "b"}]}
        >>> matches = find_all_submodels(data, Inner)
        >>> len(matches)
        2
    """
    results: list[T] = []
    if isinstance(data, dict):
        try:
            results.append(model_class.model_validate(data))
        except ValidationError:
            pass
        # Always recurse into nested values to find all matches
        for value in data.values():
            results.extend(find_all_submodels(value, model_class))
    elif isinstance(data, list):
        for item in data:
            results.extend(find_all_submodels(item, model_class))
    return results


def find_submodel(data: dict[str, Any] | list[Any], model_class: Type[T]) -> T | None:
    """Find the first occurrence of a submodel in nested JSON data.

    Recursively searches through nested dictionaries and lists to find
    the first data structure that can be validated as the given Pydantic model.

    Args:
        data: The JSON data structure (dict or list) to search.
        model_class: The Pydantic model class to search for.

    Returns:
        The first matching model instance, or None if not found.

    Example:
        >>> class Inner(BaseModel):
        ...     x: int
        >>> data = {"nested": {"x": 42}}
        >>> match = find_submodel(data, Inner)
        >>> match.x
        42
    """
    if isinstance(data, dict):
        try:
            return model_class.model_validate(data)
        except ValidationError:
            pass
        for value in data.values():
            result = find_submodel(value, model_class)
            if result:
                return result
    elif isinstance(data, list):
        for item in data:
            result = find_submodel(item, model_class)
            if result:
                return result
    return None



import math

import pytest

from pycivil.EXAGeometry.geometry import (
    Point3d,
    Vector3d,
    Vector3dApp,
    ForceCouple,
    ForceCoupleSystem
)


def test_vector3d_001():

    v1 = Vector3dApp(Point3d(1,2,3),4,5,6)
    assert v1.vx == 4
    assert v1.vy == 5
    assert v1.vz == 6
    assert Point3d(1,2,3) == v1.o

def test_force_couple_001():

    force = Vector3dApp(Point3d(0, 0, 0), (1, 1, 0))
    couple = Vector3d(0, 0, 10)

    fc = ForceCouple(force, couple)
    old_force = fc.force

    # I change origin from (0,0,0) to (1,0,0)
    fc.o = Point3d(1, 0, 0)

    # equal only if force and origin is equal
    #
    assert not fc.force == old_force

    assert fc.couple.vz == 9

def test_force_couple_002():

    force = Vector3dApp(Point3d(0.0, 0.0, 0.0), (1.0, 1.0, 0.0))
    assert force.vx == 1.0
    assert force.vy == 1.0
    assert force.vz == 0.0
    assert force.o.x == 0.0
    assert force.o.y == 0.0
    assert force.o.z == 0.0

    couple = Vector3d(0.0, 0.0, 0.0)

    force_couple = ForceCouple(force, couple)

    old_force = force_couple.force

    force_couple.o = Point3d(-1.0, 1.0, 0.0)

    # This is false because origin change from (0, 0, 0) to (-1.0, 1.0, 0.0)
    assert not force_couple.force == old_force

    assert force.norm() * force.o.distanceFrom(force_couple.o) == pytest.approx(force_couple.couple.norm())

# Two parallel forces with same distance from (0, 0, 0) give a couple of null norm.
def test_force_couple_003():
    force1 = ForceCouple(Vector3dApp(Point3d( 0.0, -1.0, 0.0), (1.0, 1.0, 0.0)), Vector3d(0.0, 0.0, 0.0))
    force2 = ForceCouple(Vector3dApp(Point3d(-1.0,  0.0, 0.0), (1.0, 1.0, 0.0)), Vector3d(0.0, 0.0, 0.0))
    resultant = force1 + force2
    assert resultant.force.norm() == 2 * math.sqrt(2)
    resultant.o = Point3d( 0.0, 0.0, 0.0)
    assert resultant.couple.norm() == 0.0

def test_force_couple_system_001():
    force1 = ForceCouple(Vector3dApp(Point3d( 0.0, -1.0, 0.0), (1.0, 1.0, 0.0)), Vector3d(0.0, 0.0, 0.0))
    force2 = ForceCouple(Vector3dApp(Point3d(-1.0,  0.0, 0.0), (1.0, 1.0, 0.0)), Vector3d(0.0, 0.0, 0.0))
    system = ForceCoupleSystem([force1, force2])
    assert system.baryOfApplicationPoints() == Point3d(-0.5, -0.5)
    reduced = system.forceCoupleResume(Point3d(0,0,0))
    assert reduced.o == Point3d(0,0,0)
    assert reduced.couple == Vector3d(0, 0, 0)
    assert reduced.force == Vector3d(2, 2, 0)







# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

"""Tests for pydmodelhelper module."""

from typing import List, Optional

from pydantic import BaseModel

from pycivil.EXAUtils.pydmodelhelper import find_all_submodels, find_submodel


class SimpleModel(BaseModel):
    x: int
    y: str


class NestedModel(BaseModel):
    name: str
    inner: SimpleModel


class DeepModel(BaseModel):
    id: int
    nested: NestedModel


class TestFindAllSubmodels:
    def test_find_in_list(self):
        """Find models nested in a list."""
        data = {"items": [{"x": 1, "y": "a"}, {"x": 2, "y": "b"}]}
        matches = find_all_submodels(data, SimpleModel)
        assert len(matches) == 2
        assert matches[0].x == 1
        assert matches[1].x == 2

    def test_find_nested_in_dict(self):
        """Find models nested inside other dicts."""
        data = {"outer": {"x": 1, "y": "a", "deeper": {"x": 2, "y": "b"}}}
        matches = find_all_submodels(data, SimpleModel)
        assert len(matches) == 2

    def test_find_with_nested_pydantic_model(self):
        """Find models that contain other Pydantic models."""
        data = {
            "wrapper": {
                "name": "parent",
                "inner": {"x": 10, "y": "nested"}
            }
        }
        matches = find_all_submodels(data, NestedModel)
        assert len(matches) == 1
        assert matches[0].name == "parent"
        assert matches[0].inner.x == 10

    def test_find_deeply_nested_models(self):
        """Find models at multiple nesting levels."""
        data = {
            "level1": {
                "id": 1,
                "nested": {
                    "name": "middle",
                    "inner": {"x": 100, "y": "deep"}
                }
            }
        }
        # Find DeepModel
        deep_matches = find_all_submodels(data, DeepModel)
        assert len(deep_matches) == 1
        assert deep_matches[0].id == 1

        # Find NestedModel (inside DeepModel)
        nested_matches = find_all_submodels(data, NestedModel)
        assert len(nested_matches) == 1
        assert nested_matches[0].name == "middle"

        # Find SimpleModel (deepest level)
        simple_matches = find_all_submodels(data, SimpleModel)
        assert len(simple_matches) == 1
        assert simple_matches[0].x == 100

    def test_find_multiple_at_different_levels(self):
        """Find multiple models at different nesting levels."""
        data = {
            "top": {"x": 1, "y": "top"},
            "container": {
                "middle": {"x": 2, "y": "middle"},
                "items": [
                    {"x": 3, "y": "item1"},
                    {"x": 4, "y": "item2"}
                ]
            }
        }
        matches = find_all_submodels(data, SimpleModel)
        assert len(matches) == 4

    def test_no_match(self):
        """Return empty list when no matches found."""
        data = {"foo": "bar", "baz": 123}
        matches = find_all_submodels(data, SimpleModel)
        assert len(matches) == 0

    def test_empty_data(self):
        """Handle empty dict."""
        matches = find_all_submodels({}, SimpleModel)
        assert len(matches) == 0

    def test_list_at_root(self):
        """Handle list at root level."""
        data = [{"x": 1, "y": "a"}, {"x": 2, "y": "b"}, {"other": "data"}]
        matches = find_all_submodels(data, SimpleModel)
        assert len(matches) == 2

    def test_complex_deep_branches(self):
        """Find models in complex structure with many deep branches."""
        data = {
            "project": {
                "metadata": {
                    "info": {"x": 1, "y": "meta"},
                    "tags": ["a", "b", "c"]
                },
                "sections": [
                    {
                        "id": "sec1",
                        "items": [
                            {"x": 2, "y": "sec1-item1"},
                            {"x": 3, "y": "sec1-item2"},
                            {
                                "nested": {
                                    "deep": {
                                        "x": 4,
                                        "y": "sec1-deep"
                                    }
                                }
                            }
                        ]
                    },
                    {
                        "id": "sec2",
                        "config": {
                            "settings": {
                                "params": {
                                    "x": 5,
                                    "y": "sec2-params"
                                }
                            }
                        },
                        "items": [
                            {"x": 6, "y": "sec2-item1"}
                        ]
                    }
                ],
                "footer": {
                    "contact": {
                        "branch1": {
                            "leaf": {"x": 7, "y": "footer-b1"}
                        },
                        "branch2": {
                            "subbranch": {
                                "leaf": {"x": 8, "y": "footer-b2"}
                            }
                        }
                    }
                }
            },
            "sidebar": {
                "widgets": [
                    {"type": "text", "data": {"x": 9, "y": "widget1"}},
                    {"type": "chart", "data": {"other": "stuff"}},
                    {
                        "type": "complex",
                        "data": {
                            "level1": {
                                "level2": {
                                    "level3": {
                                        "level4": {"x": 10, "y": "deep-widget"}
                                    }
                                }
                            }
                        }
                    }
                ]
            }
        }
        matches = find_all_submodels(data, SimpleModel)
        assert len(matches) == 10

        # Verify all values were found
        x_values = sorted([m.x for m in matches])
        assert x_values == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    def test_mixed_valid_invalid_branches(self):
        """Handle branches with partial matches and invalid data."""
        data = {
            "valid1": {"x": 1, "y": "ok"},
            "invalid1": {"x": "not_int", "y": "fail"},
            "branch": {
                "valid2": {"x": 2, "y": "ok"},
                "partial": {"x": 3},  # missing y
                "nested": {
                    "valid3": {"x": 4, "y": "ok"},
                    "invalid2": {"y": "missing_x"}
                }
            },
            "list_branch": [
                {"x": 5, "y": "ok"},
                {"invalid": True},
                [
                    {"x": 6, "y": "nested-list"},
                    {"bad": "data"}
                ]
            ]
        }
        matches = find_all_submodels(data, SimpleModel)
        assert len(matches) == 5
        x_values = sorted([m.x for m in matches])
        assert x_values == [1, 2, 4, 5, 6]


class TestFindSubmodel:
    def test_find_first_in_list(self):
        """Find first model in a list."""
        data = {"items": [{"x": 1, "y": "first"}, {"x": 2, "y": "second"}]}
        match = find_submodel(data, SimpleModel)
        assert match is not None
        assert match.y == "first"

    def test_find_nested(self):
        """Find model nested in dict."""
        data = {"wrapper": {"deep": {"x": 42, "y": "found"}}}
        match = find_submodel(data, SimpleModel)
        assert match is not None
        assert match.x == 42

    def test_find_with_nested_pydantic_model(self):
        """Find model containing another Pydantic model."""
        data = {
            "wrapper": {
                "name": "test",
                "inner": {"x": 5, "y": "inner"}
            }
        }
        match = find_submodel(data, NestedModel)
        assert match is not None
        assert match.name == "test"
        assert match.inner.x == 5

    def test_no_match_returns_none(self):
        """Return None when no match found."""
        data = {"foo": "bar"}
        match = find_submodel(data, SimpleModel)
        assert match is None

    def test_empty_data_returns_none(self):
        """Return None for empty data."""
        match = find_submodel({}, SimpleModel)
        assert match is None



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

import json
from pathlib import Path

from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import ModelInputRcSecCheck
from pycivil.EXAStructural.templateRCRect import SectionPlot
from pycivil.EXAUtils.typstReportMakers import ConcreteSectionTypstFB
from pycivil.EXAUtils.report import (
    TypstReporter,
    ReportTemplateEnum,
    getTypstTemplatesPath,
)


def _create_section_png(iData: ModelInputRcSecCheck, path: Path) -> None:
    """Create section plot PNG for Typst templates."""
    section = iData.buildRcsRectangular()
    if section is not None:
        plotter = SectionPlot(section)
        plotter.plot()
        plotter.save(path / "section.png")


def test_report_typst_section_01(tmp_path: Path):
    """Test RC section typst fragment generation with first test data."""
    with open(Path(__file__).parent / "test_report_latex_sections_01.json") as jsonFile:
        jsonObject = json.load(jsonFile)

    iData = ModelInputRcSecCheck(**jsonObject)

    section = iData.buildRcsRectangular()
    if section is not None:
        # Create section PNG for the template
        _create_section_png(iData, tmp_path)

        typstTemplatePath = getTypstTemplatesPath()
        f = ConcreteSectionTypstFB(typstTemplatePath, rcs=section).buildFragment()
        print(f.frags())
        reporter = TypstReporter(typstTemplatePath)
        reporter.linkFragments(
            template=ReportTemplateEnum.TYP_ENG_CAL, fragments=[f]
        )
        reporter.compileDocument(path=str(tmp_path), output_format="typ")
        print(f"Temporary path for test is {tmp_path}")
        assert tmp_path.joinpath("report.typ").exists()

        # Verify content has section data
        content = tmp_path.joinpath("report.typ").read_text(encoding="utf-8")
        assert "400" in content  # width or height

        # Compile to PDF
        reporter.compileDocument(path=str(tmp_path), output_format="pdf")
        assert tmp_path.joinpath("report.pdf").exists()
        assert tmp_path.joinpath("report.pdf").stat().st_size > 0


def test_report_typst_section_02(tmp_path: Path):
    """Test RC section typst fragment generation with second test data."""
    with open(Path(__file__).parent / "test_report_latex_sections_02.json") as jsonFile:
        jsonObject = json.load(jsonFile)

    iData = ModelInputRcSecCheck(**jsonObject)

    section = iData.buildRcsRectangular()
    if section is not None:
        # Create section PNG for the template
        _create_section_png(iData, tmp_path)

        typstTemplatePath = getTypstTemplatesPath()
        f = ConcreteSectionTypstFB(typstTemplatePath, rcs=section).buildFragment()
        print(f.frags())
        reporter = TypstReporter(typstTemplatePath)
        reporter.linkFragments(
            template=ReportTemplateEnum.TYP_ENG_CAL, fragments=[f]
        )
        reporter.compileDocument(path=str(tmp_path), output_format="pdf")
        print(f"Temporary path for test is {tmp_path}")
        assert tmp_path.joinpath("report.typ").exists()
        assert tmp_path.joinpath("report.pdf").exists()
        assert tmp_path.joinpath("report.pdf").stat().st_size > 0



import pytest

from pycivil.EXAStructuralModel import fe

# Test that show difference between model and mesh assigning relative tags
# The way to force mesh recognition elements with tags is assign a physical
# group during creation of 0d and 1d element
def test_fe_1d_elements_001():
    # It builds an analytical model
    #
    am = fe.FEModel(descr=f"Model for {__name__}", modelName=__name__)

    # Only one model created currently by constructor ...
    #
    assert len(am.getModels()) == 1

    # ... named as passed in modelName ard during init
    #
    assert am.getModels()[0] == __name__

    # I build 3 nodes, forcing tag to specific values to 3, 5, 7
    #
    node_3 = am.addNode(
        x=0.0, y=0.0, z=5.0,
        tag=3, group=True, tagGroup=13
    )
    assert node_3 == 3
    node_3_pg = am.getNodesByPhysicalGroup(13)
    assert node_3_pg[0] == 3

    node_5 = am.addNode(
        x=5.0, y=0.0, z=5.0,
        tag=5, group=True, tagGroup=15
    )
    assert node_5 == 5
    node_5_pg = am.getNodesByPhysicalGroup(15)
    assert node_5_pg[0] == 5

    node_7 = am.addNode(
        x=10.0, y=0.0, z=5.0,
        tag=7, group=True, tagGroup=17
    )
    assert node_7 == 7
    node_7_pg = am.getNodesByPhysicalGroup(17)
    assert node_7_pg[0] == 7

    # I build 2 frames linked to nodes meshed in one element each one forcing
    # tag to 1 and 2.
    #
    frame_20 = am.addFrame(
        tagStart=node_5, tagEnd=node_7, tag=20,
        group=True, tagGroup=21, nb=1
    )
    assert frame_20 == 20

    frame_20_pg = am.getFramesByPhysicalGroup(21)
    assert frame_20_pg[0] == 20

    frame_10 = am.addFrame(
        tagStart=node_3, tagEnd=node_5, tag=10,
        group=True, tagGroup=11, nb=1
    )
    assert frame_10 == 10

    frame_10_pg = am.getFramesByPhysicalGroup(11)
    assert frame_10_pg[0] == 10

    # am.show()




# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

"""Tests for Typst cheat sheets generation.

These tests verify the Typst and PDF cheat sheet generation for:
- Concrete material properties (NTC2018)
- Rebar material properties (NTC2018)
- Beam minimum area (NTC2018)
- Plate minimum area (EC2-1-1)
"""

from pathlib import Path

from pycivil.EXAStructural.cheatsheets.codeEC2Rules import (
    PlateMinimumArea,
    PlateMinimumAreaInput,
    PlateMinimumAreaLogs,
    SolverPlateMinRebar,
)
from pycivil.EXAStructural.cheatsheets.codeNTC2018Rules import (
    BeamMinimumArea,
    BeamMinimumAreaInput,
    BeamMinimumAreaLogs,
    BeamMinimumAreaOutput,
    ConcreteMaterial,
    ConcreteMaterialInput,
    ConcreteMaterialOutput,
    RebarMaterial,
    RebarMaterialInput,
    RebarMaterialOutput,
    SolverBeamMinRebar,
    SolverConcrete,
    SolverSteelRebar,
)
from pycivil.EXAUtils import typstCheatSheets
from pycivil.EXAUtils.report import (
    TypstFragment,
    TypstReporter,
    ReportTemplateEnum,
    getTypstTemplatesPath,
)


def test_concrete_material_typst(tmp_path: Path):
    """Test concrete material cheat sheet generation in Typst format."""
    matInput = ConcreteMaterialInput()
    solver = SolverConcrete(inputModel=matInput)

    assert solver.run()

    outputData = solver.getModelOutput()
    assert isinstance(outputData, ConcreteMaterialOutput)
    material = ConcreteMaterial(inputData=matInput, outputData=outputData)

    typstTemplatePath = getTypstTemplatesPath()

    fragBuilder = typstCheatSheets.ConcreteMaterialTypstCS(
        typstTemplatePath=typstTemplatePath, data=material
    )

    frag_title = TypstFragment(typstTemplatePath)
    frag_title.add(line="== Calcestruzzo adottato per l'elemento")
    frag = fragBuilder.buildFragment()

    reporter = TypstReporter(typstTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.TYP_ENG_CAL, fragments=[frag_title, frag]
    )
    reporter.compileDocument(path=str(tmp_path), fileName="TYP_CS_CONCRETE", output_format="pdf")
    assert tmp_path.joinpath("TYP_CS_CONCRETE.typ").exists()
    assert tmp_path.joinpath("TYP_CS_CONCRETE.pdf").exists()
    assert tmp_path.joinpath("TYP_CS_CONCRETE.pdf").stat().st_size > 0


def test_rebar_material_typst(tmp_path: Path):
    """Test rebar material cheat sheet generation in Typst format."""
    matInput = RebarMaterialInput()
    solver = SolverSteelRebar(inputModel=matInput)

    assert solver.run()
    outputData = solver.getModelOutput()
    assert isinstance(outputData, RebarMaterialOutput)
    material = RebarMaterial(inputData=matInput, outputData=outputData)

    typstTemplatePath = getTypstTemplatesPath()
    fragBuilder = typstCheatSheets.RebarMaterialTypstCS(
        typstTemplatePath=typstTemplatePath, data=material
    )
    frag_title = TypstFragment(typstTemplatePath)
    frag_title.add(line="== Acciaio adottato per l'elemento")
    frag = fragBuilder.buildFragment()

    reporter = TypstReporter(typstTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.TYP_ENG_CAL, fragments=[frag_title, frag]
    )
    reporter.compileDocument(path=str(tmp_path), fileName="TYP_CS_REBAR", output_format="pdf")
    assert tmp_path.joinpath("TYP_CS_REBAR.typ").exists()
    assert tmp_path.joinpath("TYP_CS_REBAR.pdf").exists()
    assert tmp_path.joinpath("TYP_CS_REBAR.pdf").stat().st_size > 0


def test_beam_minimum_area_typst(tmp_path: Path):
    """Test beam minimum area cheat sheet generation in Typst format."""
    dataInput = BeamMinimumAreaInput()
    solver = SolverBeamMinRebar(inputModel=dataInput)

    assert solver.run()

    logsData = solver.getModelLogs()
    assert isinstance(logsData, BeamMinimumAreaLogs)

    outputData = solver.getModelOutput()
    assert isinstance(outputData, BeamMinimumAreaOutput)

    data = BeamMinimumArea(
        inputData=dataInput,
        logsData=logsData,
        outputData=outputData,
    )

    typstTemplatePath = getTypstTemplatesPath()

    fragBuilder = typstCheatSheets.BeamMinimumAreaTypstCS(
        typstTemplatePath=typstTemplatePath, data=data
    )

    frag_title = TypstFragment(typstTemplatePath)
    frag_title.add(line="== Area minima delle travi secondo NTC2018")
    frag = fragBuilder.buildFragment()

    reporter = TypstReporter(typstTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.TYP_ENG_CAL, fragments=[frag_title, frag]
    )
    reporter.compileDocument(path=str(tmp_path), fileName="TYP_CS_BEAM", output_format="pdf")
    assert tmp_path.joinpath("TYP_CS_BEAM.typ").exists()
    assert tmp_path.joinpath("TYP_CS_BEAM.pdf").exists()
    assert tmp_path.joinpath("TYP_CS_BEAM.pdf").stat().st_size > 0


def test_plate_minimum_area_typst(tmp_path: Path):
    """Test plate minimum area cheat sheet generation in Typst format."""
    dataInput = PlateMinimumAreaInput()
    solver = SolverPlateMinRebar(inputModel=dataInput)

    assert solver.run()

    logsData = solver.getModelLogs()
    assert isinstance(logsData, PlateMinimumAreaLogs)

    data = PlateMinimumArea(
        inputData=dataInput,
        logsData=logsData,
    )

    typstTemplatePath = getTypstTemplatesPath()

    fragBuilder = typstCheatSheets.PlateMinimumAreaTypstCS(
        typstTemplatePath=typstTemplatePath, data=data
    )

    frag_title = TypstFragment(typstTemplatePath)
    frag_title.add(line="== Area minima delle piastre secondo EC2")
    frag = fragBuilder.buildFragment()

    reporter = TypstReporter(typstTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.TYP_ENG_CAL, fragments=[frag_title, frag]
    )
    reporter.compileDocument(path=str(tmp_path), fileName="TYP_CS_PLATE", output_format="pdf")
    assert tmp_path.joinpath("TYP_CS_PLATE.typ").exists()
    assert tmp_path.joinpath("TYP_CS_PLATE.pdf").exists()
    assert tmp_path.joinpath("TYP_CS_PLATE.pdf").stat().st_size > 0



import numpy as np
from enum import Enum
from typing import Tuple, Dict, List
from numpy.typing import NDArray

from matplotlib import (
    pyplot as plt,
)
from matplotlib.pyplot import Figure, Axes
from matplotlib.lines import Line2D
from matplotlib.patches import Circle


from pycivil.EXAGeometry.geometry import Point2d


TpLine2D = Tuple[NDArray[np.double], NDArray[np.double]]

def twoP2N(p1: Point2d, p2: Point2d) -> TpLine2D:
    return np.array([p1.x, p2.x]), np.array([p1.y, p2.y])

class Basic2DType(Enum):
    LINE2D = 1,
    CIRCLE2D = 2,
    RECTANGLE2D = 3,

class Plotter:
    def __init__(self, axes: Axes) -> None:
        self.__id: int = 0
        self.__axes: Axes = axes
        self.__obj_tp: Dict[int, Basic2DType] = {}
        self.__lines: Dict[int, Tuple[Point2d, Point2d]]= {}
        self.__circles: Dict[int, Tuple[Point2d, float]]= {}

    def __new_id(self) -> int:
        self.__id += 1
        return self.__id

    def add_line(self, p1: Point2d, p2: Point2d) -> int:
        new_id = self.__new_id()
        self.__lines[new_id] = (p1, p2)
        self.__obj_tp[new_id] = Basic2DType.LINE2D
        return new_id

    def add_circle(self, center: Point2d, radius: float) -> int:
        new_id = self.__new_id()
        self.__circles[new_id] = (center, radius)
        self.__obj_tp[new_id] = Basic2DType.CIRCLE2D
        return new_id

    def rotate(self, items: List[int], angle: float, center: Point2d) -> None:
        for i in items:
            if self.__obj_tp.get(i) == Basic2DType.LINE2D:
                points = self.__lines.get(i)
                assert points is not None
                points[0].rotate(angle, center)
                points[1].rotate(angle, center)
            elif self.__obj_tp.get(i) == Basic2DType.CIRCLE2D:
                circle = self.__circles.get(i)
                assert circle is not None
                circle[0].rotate(angle, center)
            else:
                raise ValueError(f"Object with id={id} is unknown type !!!")

    def plot(self) -> Axes:

        for k in self.__lines.keys():
            np_data = twoP2N(self.__lines[k][0], self.__lines[k][1])
            self.__axes.add_line(Line2D(np_data[0], np_data[1]))

        for k in self.__circles.keys():
            self.__axes.add_artist(
                Circle(
                    (self.__circles[k][0].x,self.__circles[k][0].y),
                    self.__circles[k][1], fill=False),
            )

        return self.__axes

# EXAMPLE of use
# fig: Figure = plt.figure()
# ax: Axes = fig.add_subplot(1, 1, 1)
# ax.set_aspect("equal", "datalim")
# pl = Plotter(ax)
# #################################
# #           GEOM OP
# ################################
# l1 = pl.add_line(Point2d(0.0, 0.0), Point2d(1.0, 1.0))
# l2 = pl.add_line(Point2d(1.0, 0.0), Point2d(0.0, 1.0))
# c1 = pl.add_circle(Point2d(0.5, 0.5), 0.25)
# pl.rotate([l1, l2], 45, Point2d(0.5, 0.5))
# pl.rotate([c1], 45, Point2d(0.0, 0.0))
# #################################
# plt.autoscale()
# pl.plot()
#
#
# plt.show()



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

"""Full document generation tests for typst reports."""

import json
from pathlib import Path

from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.loads import ForcesOnSection
from pycivil.EXAStructural.loads import Frequency_Enum as Fr
from pycivil.EXAStructural.loads import LimiteState_Enum as Ls
from pycivil.EXAStructural.materials import Concrete, ConcreteSteel
from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import ModelInputRcSecCheck
from pycivil.EXAStructural.templateRCRect import SectionPlot
from pycivil.EXAUtils.typstReportMakers import (
    ConcreteTypstFB,
    ConcreteSectionTypstFB,
    ForcesOnSectionListTypstFB,
    SteelConcreteTypstFB,
)
from pycivil.EXAUtils.report import (
    EnumFBSection,
    TypstReporter,
    ReportProperties,
    ReportTemplateEnum,
    getTypstTemplatesPath,
)


def _create_section_png(iData: ModelInputRcSecCheck, path: Path) -> None:
    """Create section plot PNG for Typst templates."""
    section = iData.buildRcsRectangular()
    if section is not None:
        plotter = SectionPlot(section)
        plotter.plot()
        plotter.save(path / "section.png")


def test_report_typst_full_document(tmp_path: Path):
    """Test full engineering calculation document generation."""
    code = Code("NTC2018")
    typstTemplatePath = getTypstTemplatesPath()

    # Materials
    concrete = Concrete()
    concrete.setByCode(codeObj=code, catstr="C25/30")
    concreteFB = ConcreteTypstFB(typstTemplatePath, concrete)
    concreteFB.setFragmentOptions({
        "section_title": "Materiale: Calcestruzzo",
        "section_level": EnumFBSection.SEC_SECTION
    })
    f_concrete = concreteFB.buildFragment()

    steel = ConcreteSteel()
    steel.setByCode(code, "B450C")
    steelFB = SteelConcreteTypstFB(typstTemplatePath, steel)
    steelFB.setFragmentOptions({
        "section_title": "Materiale: Acciaio",
        "section_level": EnumFBSection.SEC_SECTION
    })
    f_steel = steelFB.buildFragment()

    # Section
    with open(Path(__file__).parent / "test_report_latex_sections_01.json") as jsonFile:
        jsonObject = json.load(jsonFile)
    iData = ModelInputRcSecCheck(**jsonObject)
    section = iData.buildRcsRectangular()

    f_section = None
    if section is not None:
        # Create section PNG
        _create_section_png(iData, tmp_path)

        sectionFB = ConcreteSectionTypstFB(typstTemplatePath, rcs=section)
        sectionFB.setFragmentOptions({
            "section_title": "Geometria della sezione",
            "section_level": EnumFBSection.SEC_SECTION
        })
        f_section = sectionFB.buildFragment()

    # Loads
    loadsFB = ForcesOnSectionListTypstFB(typstTemplatePath)
    loadsFB.setFragmentOptions({
        "section_title": "Combinazioni di carico",
        "section_level": EnumFBSection.SEC_SECTION
    })
    listOfForces = loadsFB.forces()
    listOfForces.append(
        ForcesOnSection(
            Fx=0, Fy=0, Fz=1e3 * 500,
            Mx=0, My=1e3 * 150000, Mz=0,
            descr="SLU - Combinazione 1",
            id=1,
            limitState=Ls.ULTIMATE,
            frequency=Fr.FREQUENCY_ND,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=0, Fy=0, Fz=1e3 * 350,
            Mx=0, My=1e3 * 100000, Mz=0,
            descr="SLE - Rara",
            id=2,
            limitState=Ls.SERVICEABILITY,
            frequency=Fr.CHARACTERISTIC,
        )
    )
    listOfForces.append(
        ForcesOnSection(
            Fx=0, Fy=0, Fz=1e3 * 300,
            Mx=0, My=1e3 * 80000, Mz=0,
            descr="SLE - Frequente",
            id=3,
            limitState=Ls.SERVICEABILITY,
            frequency=Fr.FREQUENT,
        )
    )
    f_loads = loadsFB.buildFragment()

    # Assemble fragments
    fragments = [f_concrete, f_steel]
    if f_section is not None:
        fragments.append(f_section)
    fragments.append(f_loads)

    # Create reporter with properties
    reporter = TypstReporter(typstTemplatePath)
    reporter.setProperties(ReportProperties(
        project_brief="Test Project - RC Beam Analysis",
        module_name="PyCivil",
        module_version="0.2.20",
        report_designer="Test Engineer",
        report_date="2026-02-18",
        report_time="10:30",
        report_token="TEST-001",
    ))
    reporter.linkFragments(
        template=ReportTemplateEnum.TYP_ENG_CAL,
        fragments=fragments
    )
    reporter.compileDocument(path=str(tmp_path), fileName="full_report", output_format="pdf")

    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("full_report.typ").exists()
    assert tmp_path.joinpath("full_report.pdf").exists()
    assert tmp_path.joinpath("full_report.pdf").stat().st_size > 0

    # Verify content
    content = tmp_path.joinpath("full_report.typ").read_text(encoding="utf-8")
    assert "Test Project" in content
    assert "C25/30" in content or "25.0" in content
    assert "B450" in content or "450" in content


def _make_dummy_logo(path: Path) -> None:
    """Generate a dummy PNG logo with matplotlib, similar in style to the PyCivil logo."""
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches

    fig, ax = plt.subplots(figsize=(6.4, 1.6))
    fig.patch.set_facecolor("white")
    ax.set_facecolor("white")
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.axis("off")

    red = "#a3202a"
    rect = mpatches.FancyBboxPatch(
        (0.02, 0.1), 0.96, 0.85,
        boxstyle="round,pad=0.02",
        linewidth=0,
        facecolor=red,
    )
    ax.add_patch(rect)
    ax.text(
        0.5, 0.55, "LiviCyP",
        ha="center", va="center",
        fontsize=28, fontweight="bold", color="white",
        transform=ax.transAxes,
    )
    ax.text(
        0.5, 0.18, "only for free Structural Engineers",
        ha="center", va="center",
        fontsize=8, color="white",
        transform=ax.transAxes,
    )

    fig.savefig(path, dpi=96, bbox_inches="tight", pad_inches=0.05)
    plt.close(fig)


def test_report_typst_custom_logo(tmp_path: Path):
    """Test that a custom logo is copied to output and used in the report."""
    code = Code("NTC2018")
    typstTemplatePath = getTypstTemplatesPath()

    # Generate a dummy logo with matplotlib
    custom_logo_dir = tmp_path / "assets"
    custom_logo_dir.mkdir()
    custom_logo = custom_logo_dir / "company_logo.png"
    _make_dummy_logo(custom_logo)

    concrete = Concrete()
    concrete.setByCode(codeObj=code, catstr="C25/30")
    concreteFB = ConcreteTypstFB(typstTemplatePath, concrete)

    output_dir = tmp_path / "output"
    output_dir.mkdir()

    reporter = TypstReporter(typstTemplatePath)
    reporter.setProperties(ReportProperties(
        project_brief="Custom Logo Test",
        module_name="PyCivil",
        module_version="0.2.20",
        report_designer="Test Engineer",
        report_date="2026-02-18",
        report_time="10:30",
        report_token="TEST-LOGO",
        report_logo=custom_logo,
    ))
    reporter.linkFragments(
        template=ReportTemplateEnum.TYP_ENG_CAL,
        fragments=[concreteFB.buildFragment()]
    )
    reporter.compileDocument(path=str(output_dir), fileName="logo_report", output_format="pdf")

    assert output_dir.joinpath("logo_report.typ").exists()
    assert output_dir.joinpath("logo_report.pdf").exists()
    assert output_dir.joinpath("logo_report.pdf").stat().st_size > 0
    # Logo must be copied alongside the .typ file
    assert output_dir.joinpath("company_logo.png").exists()
    # Verify the .typ file references the custom logo name
    content = output_dir.joinpath("logo_report.typ").read_text(encoding="utf-8")
    assert "company_logo.png" in content


def test_report_typst_builder_api(tmp_path: Path):
    """Test using builder API instead of fragments list."""
    code = Code("NTC2018")
    typstTemplatePath = getTypstTemplatesPath()

    concrete = Concrete()
    concrete.setByCode(codeObj=code, catstr="C25/30")
    concreteFB = ConcreteTypstFB(typstTemplatePath, concrete)

    steel = ConcreteSteel()
    steel.setByCode(code, "B450C")
    steelFB = SteelConcreteTypstFB(typstTemplatePath, steel)

    reporter = TypstReporter(typstTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.TYP_ENG_CAL,
        builder=[concreteFB, steelFB]
    )
    reporter.compileDocument(path=str(tmp_path), fileName="builder_report", output_format="pdf")

    assert tmp_path.joinpath("builder_report.typ").exists()
    assert tmp_path.joinpath("builder_report.pdf").exists()
    assert tmp_path.joinpath("builder_report.pdf").stat().st_size > 0



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

"""Module with classes for generate reports in various formats.

Main classes in module are Reporter (for LaTeX), MarkdownReporter (for Markdown),
and TypstReporter (for Typst). Reporter uses LaTeX executable for building reports,
MarkdownReporter uses weasyprint for PDF generation, and TypstReporter uses the
typst Python package for PDF generation.

Classes list:

    1. EnumFBSection
    2. Fragment
    3. FragmentsBuilder
    4. getTemplatesPath
    5. getMarkdownTemplatesPath
    6. getTypstTemplatesPath
    7. MarkdownFragment
    8. MarkdownFragmentsBuilder
    9. MarkdownReporter
    10. ReportDriverEnum
    11. Reporter
    12. ReportProperties
    13. ReportTemplateEnum
    14. TypstFragment
    15. TypstFragmentsBuilder
    16. TypstReporter
"""

import os
import subprocess
from enum import Enum, unique
from importlib.resources import files
from pathlib import Path
from typing import Any, Dict, List, Literal, Union
import shutil

from jinja2 import Environment, FileSystemLoader
from pydantic import BaseModel

from pycivil.EXAUtils.EXAExceptions import EXAExceptions
from pycivil.EXAUtils import logging as logger


@unique
class ReportDriverEnum(Enum):
    PDFLATEX = 1
    WEASYPRINT = 2  # Markdown → HTML → PDF
    TYPST = 3  # Typst → PDF


@unique
class ReportTemplateEnum(Enum):
    """This Enum describe the main documents level

    Attributes:
        TEX_ENG_CAL: style for official reports
        TEX_KOMA: styled with KOMA scripts LaTex
        TEX_MAIN: styled with main LaTex
        MD_ENG_CAL: markdown style for official reports
        MD_MAIN: markdown styled with main template
    """
    TEX_ENG_CAL = 1
    TEX_KOMA = 2
    TEX_MAIN = 3
    MD_ENG_CAL = 4
    MD_MAIN = 5
    TYP_ENG_CAL = 6
    TYP_MAIN = 7


def getTemplatesPath() -> Path:
    """Funtion to get templates latex path.

    Returns:
        A Path object.
    """
    return Path(str(files("pycivil") / "templates" / "latex"))


def getMarkdownTemplatesPath() -> Path:
    """Function to get templates markdown path.

    Returns:
        A Path object.
    """
    return Path(str(files("pycivil") / "templates" / "markdown"))


def getTypstTemplatesPath() -> Path:
    """Function to get templates typst path.

    Returns:
        A Path object.
    """
    return Path(str(files("pycivil") / "templates" / "typst"))


class Fragment:
    """Class that hold fragments as list of strings for use with Reporter.

    This class is a container for complete a template (ex. LaTex template).
    Template are processed after with Reporter class. Will be generated
    automatically a type of file. For example PDF or MD etc.

    """
    def __init__(self, latexTemplatePath: Path = Path()):
        self.__fragment: List[str] = []
        self.__latexTemplatePath: Path = latexTemplatePath
        self.__logoFilePath: Path = Path(self.__latexTemplatePath) / Path("logo.png")

    def set_logo(self, logo_path: Path) -> None:
        self.__logoFilePath = logo_path

    def add_line(self, line: str) -> None:
        self.__fragment.append(line)

    def add_lines(self, lines: List[str]) -> None:
        for ll in lines:
            if ll is not None:
                self.__fragment.append(ll)

    def add_template(self, name: str, placeholders: Dict[str, Any]) -> None:
        file_loader = FileSystemLoader(searchpath=self.__latexTemplatePath)
        env = Environment(
            block_start_string=r"\BLOCK{",
            block_end_string="}",
            variable_start_string=r"\VAR{",
            variable_end_string="}",
            comment_start_string=r"\#{",
            comment_end_string="}",
            line_statement_prefix="%-",
            line_comment_prefix="%#",
            trim_blocks=True,
            autoescape=False,
            loader=file_loader,
        )
        jtemplate = env.get_template(name)
        rendered = jtemplate.render(placeholders, logo_path=self.__logoFilePath.as_posix())
        self.__fragment.append(rendered)

    def add(
        self,
        line: Union[str, None] = None,
        lines: Union[List[str], None] = None,
        templateName: Union[str, None] = None,
        templatePlaceholders: Union[Dict[str, Any], None] = None,
    ) -> int:
        if (
            line is not None
            and templateName is None
            and templatePlaceholders is None
            and lines is None
        ):
            self.add_line(line)
        elif (
            line is None
            and lines is None
            and templateName is not None
            and templatePlaceholders is not None
        ):
            self.add_template(templateName, templatePlaceholders)
        elif (
            line is None
            and lines is not None
            and templateName is None
            and templatePlaceholders is None
        ):
            self.add_lines(lines)
        else:
            raise EXAExceptions("0001", "Wrong args", "")
        return len(self.__fragment)

    def frags(self) -> List[str]:
        return self.__fragment


@unique
class EnumFBSection(Enum):
    SEC_CHAPTER = 1
    SEC_SECTION = 2
    SEC_SUBSECTION = 3
    SEC_SUBSUBSECTION = 4


class FragmentsBuilder:
    def __init__(self):
        self.__fragment: Union[Fragment, None] = None
        self.__fragment_options: Dict[str, Any] = {}

    def _setFragment(self, f: Fragment) -> None:
        self.__fragment = f

    def getFragmentOptions(self) -> Dict[str, Any]:
        return self.__fragment_options

    def _setFragmentOptions(self, options: Dict[str, Any]) -> bool:
        self.__fragment_options = options
        return True

    def setFragmentOptions(self, options: Dict[str, Any]) -> bool:
        for name_opt, value_opt in options.items():
            if name_opt not in self.__fragment_options.keys():
                raise ValueError(f"{name_opt} is not a valid option")
        self.__fragment_options = options
        return True

    def buildFragment(self) -> Fragment:
        raise NotImplementedError("Need to be implemented")

    def fragment(self) -> Union[Fragment, None]:
        return self.__fragment


class MarkdownFragment:
    """Class that holds fragments as list of strings for use with MarkdownReporter.

    This class is a container for complete a markdown template.
    Templates are processed using standard Jinja2 delimiters ({{ }}, {% %}).
    """
    def __init__(self, markdownTemplatePath: Path = Path()):
        self.__fragment: List[str] = []
        self.__markdownTemplatePath: Path = markdownTemplatePath
        self.__logoFilePath: Path = Path(self.__markdownTemplatePath) / Path("logo.png")

    def set_logo(self, logo_path: Path) -> None:
        self.__logoFilePath = logo_path

    def add_line(self, line: str) -> None:
        self.__fragment.append(line)

    def add_lines(self, lines: List[str]) -> None:
        for ll in lines:
            if ll is not None:
                self.__fragment.append(ll)

    def add_template(self, name: str, placeholders: Dict[str, Any]) -> None:
        """Add a rendered template using standard Jinja2 delimiters."""
        file_loader = FileSystemLoader(searchpath=self.__markdownTemplatePath)
        env = Environment(
            trim_blocks=True,
            autoescape=False,
            loader=file_loader,
        )
        jtemplate = env.get_template(name)
        rendered = jtemplate.render(placeholders, logo_path=self.__logoFilePath.as_posix())
        self.__fragment.append(rendered)

    def add(
        self,
        line: Union[str, None] = None,
        lines: Union[List[str], None] = None,
        templateName: Union[str, None] = None,
        templatePlaceholders: Union[Dict[str, Any], None] = None,
    ) -> int:
        if (
            line is not None
            and templateName is None
            and templatePlaceholders is None
            and lines is None
        ):
            self.add_line(line)
        elif (
            line is None
            and lines is None
            and templateName is not None
            and templatePlaceholders is not None
        ):
            self.add_template(templateName, templatePlaceholders)
        elif (
            line is None
            and lines is not None
            and templateName is None
            and templatePlaceholders is None
        ):
            self.add_lines(lines)
        else:
            raise EXAExceptions("0001", "Wrong args", "")
        return len(self.__fragment)

    def frags(self) -> List[str]:
        return self.__fragment


class MarkdownFragmentsBuilder:
    """Abstract base class for markdown fragment builders."""
    def __init__(self) -> None:
        self.__fragment: Union[MarkdownFragment, None] = None
        self.__fragment_options: Dict[str, Any] = {}

    def _setFragment(self, f: MarkdownFragment) -> None:
        self.__fragment = f

    def getFragmentOptions(self) -> Dict[str, Any]:
        return self.__fragment_options

    def _setFragmentOptions(self, options: Dict[str, Any]) -> bool:
        self.__fragment_options = options
        return True

    def setFragmentOptions(self, options: Dict[str, Any]) -> bool:
        for name_opt, value_opt in options.items():
            if name_opt not in self.__fragment_options.keys():
                raise ValueError(f"{name_opt} is not a valid option")
        self.__fragment_options = options
        return True

    def buildFragment(self) -> MarkdownFragment:
        raise NotImplementedError("Need to be implemented")

    def fragment(self) -> Union[MarkdownFragment, None]:
        return self.__fragment


class ReportProperties(BaseModel):
    project_brief: str = "Brief"
    module_name: str = "Module Name"
    module_version: str = "Module Version"
    report_designer: str = "Designer user name"
    report_date: str = "18/01/1972"
    report_time: str = "01:30"
    report_token: str = "xxxxxxx"
    report_logo: Path | None = None


class Reporter:
    r"""Class that compiles fragments to produce PDF or others.

    This class compiles fragments to produce PDF or others type of files.
    Actually produces PDF using LaTex files by one of LaTeX implementations as
    MikTex (popular implementation for Windows) or TexLive (popular on Linux).

    Typical usage example:

    ```
     reporter = Reporter(latexTemplatePath)
     reporter.linkFragments(ReportDriverEnum.PDFLATEX, ReportTemplateEnum.TEX_ENG_CAL, [frag_title, frag])
     reporter.makePDF(path=str(working_path), fileName="TEX_ENG_CAL")
    ```
    """
    def __init__(self, latexTemplatePath: Path):
        self.__main_file_name: str = ""
        self.__templated: Union[str, None] = ""
        self.__sepForTex: str = "\n"
        self.__opt_makeGlossary: bool = False
        self.__ll: Literal[0, 1, 2, 3] = 3
        self.__latexTemplatePath: Path = latexTemplatePath
        self.__properties: ReportProperties = ReportProperties()

    def setProperties(self, prop: ReportProperties) -> None:
        self.__properties = prop

    def compileDocument(
        self,
        path: str = "",
        fileName: str = "report",
        verbose: bool = False,
        deleteIfExists: bool = True,
    ) -> None:
        """Build a PDF using driver LATEXPDF

        Write before templated in <path> with name <fileName>.
        Extension are automatically added at the end of file name.

        Args:
            deleteIfExists (bool, optional): if True delete file PDF if exists
            verbose (bool, optional):
            path (str, optional): where intermediate file are crated. Defaults to ''.
            fileName (str, optional): input file name. Defaults to 'report'.
        """
        completeFileName = Path(f"{path}/{fileName}")
        completeFileNameTex = Path(f"{completeFileName}.tex")
        completeFileNamePdf = Path(f"{completeFileName}.pdf")

        # Save tex file
        with open(file=completeFileNameTex, mode="w", encoding="utf-8") as f:
            f.write(self.__templated)  # type: ignore

        if deleteIfExists:
            if os.path.exists(completeFileNamePdf):
                os.remove(completeFileNamePdf)
                logger.log(
                    tp="INF", level=self.__ll, msg=f"removed file {completeFileNamePdf}"
                )
            else:
                logger.log(
                    tp="INF",
                    level=self.__ll,
                    msg=f"file do not exists {completeFileNamePdf}",
                )

        # Compile
        os.chdir(path)
        os.listdir()
        #
        # FIRST RUN PDFLATEX
        #
        print("Setted make glossary:")
        print("... 1/4 run first pdflatex")
        self._run_latex(str(completeFileNameTex), verbose)
        if not self.__opt_makeGlossary:
            print("... 2/4 no run makeglossaries")
        else:
            self._make_glossaries(fileName, path, verbose)

        #
        # SECOND RUN PDFLATEX
        #
        print("... 3/4 run pdflatex for tables width adjust")
        self._run_latex(str(completeFileNameTex), verbose)

        #
        # THIRD RUN PDFLATEX
        #
        print("... 4/4 run pdflatex for tables width adjust")
        if self._run_latex(str(completeFileNameTex), verbose):
            print(f"PDF generated as {completeFileNamePdf}")

    def _make_glossaries(self, fileName, path, verbose):
        #
        # SECOND RUN MAKEGLOSSARIES
        #
        print("... 2/4 run makeglossaries")
        print(f"Current working directory: {os.getcwd()}")
        cwdOld = os.getcwd()
        os.chdir(path)
        print(f"for makeglossaries working directory: {os.getcwd()}")
        self._run_command(["makeglossaries", fileName], verbose)
        os.chdir(cwdOld)

    def _run_latex(self, file_name: str, verbose: bool = False) -> bool:
        logger.log("INF", msg=f"File name pdflatex run is in {file_name}", level=self.__ll)

        return self._run_command(["pdflatex", "--no-shell-escape",file_name], verbose)

    @staticmethod
    def _run_command(args: List[str], verbose: bool = False) -> bool:
        # Check if pdflatex is in PATH

        program_path = shutil.which(args[0])
        if shutil.which(args[0]):
            print(f"{program_path} found in PATH {program_path}")
        else:
            print(f"{program_path} NOT found in PATH - need to install")
        print(args)

        if os.name == "nt":
            x = subprocess.run(
                executable=None,
                args=args,
                text=True,
                shell=True,
                capture_output=not verbose)
        else:
            x = subprocess.run(
                executable=args[0],
                args=args[1:len(args)],
                text=True,
                shell=True,
                capture_output=not verbose)

        if x.returncode != 0:
            print("... Exit-code not 0, check result!")
        else:
            print("... Exit-code 0")
        return x.returncode == 0

    def linkFragments(
        self,
        template: ReportTemplateEnum,
        driver: ReportDriverEnum = ReportDriverEnum.PDFLATEX,
        fragments: List[Fragment] | None = None,
        builder: Union[FragmentsBuilder, List[FragmentsBuilder]] | None = None,
        glossary: bool = False,
        main_file_name: str = "",
    ) -> bool:
        """Links fragments adding also a main documents.

        Links fragments adding also a main documents. For example in LaTex
        the linking process add

        ```LaTex
            \\documentclass[11pt, a4paper]{article}
            \\usepackage[latin1]{inputenc}
            ...
            \\begin{document}
            ...
            Fragment line 1
            Fragment line 2
            ...
            Fragment line n
            ...
            \\end{document}
        ```

        Args:
            template (ReportTemplateEnum):
             Enum for main document template.
            driver (ReportDriverEnum, Optional):
             Leave default option.
            fragments (List[Fragment], Optional):
             List of rows to put in main document.
            builder (Union[FragmentsBuilder, List[FragmentsBuilder]], Optional):
             List of fragment builder. This is alternative to fragments
             argument. Default is None.
            glossary (bool, optional): Default is False.
            main_file_name (str, optional): Default is "".

        Returns:
            True on success, False otherwise
        """
        self.__opt_makeGlossary = glossary
        self.__main_file_name = main_file_name
        print("buildPDF: start ...")

        if builder is not None:
            if not isinstance(builder, list):
                builders = [builder]
            else:
                builders = builder

            fragments = []
            for b in builders:
                fragments.append(b.buildFragment())

        united = []
        assert fragments is not None
        for f in fragments:
            if f.frags() is not None:
                united += f.frags()  # type: ignore
            else:
                print("ERR: None fragment")
                print("buildPDF: Quit")
                return False

        if ReportDriverEnum.PDFLATEX != driver:
            print("ERR: driver unknown !!! quit")
            return False

        file_loader = FileSystemLoader(searchpath=self.__latexTemplatePath)
        env = Environment(
            block_start_string="\\BLOCK{",
            block_end_string="}",
            variable_start_string="\\VAR{",
            variable_end_string="}",
            comment_start_string="\\#{",
            comment_end_string="}",
            line_statement_prefix="%-",
            line_comment_prefix="%#",
            trim_blocks=True,
            autoescape=False,
            loader=file_loader,
        )

        print("build with pdflatex ...")

        # LaTex substitution strings
        #
        # 1) Changing "_" with "\_" for latex
        project_brief = self.__properties.project_brief.replace("_", r"\_")

        if (
            template == ReportTemplateEnum.TEX_ENG_CAL
            or template == ReportTemplateEnum.TEX_MAIN
        ):
            print("... template TEX_ENG_CAL")
            if template == ReportTemplateEnum.TEX_ENG_CAL:
                jtemplate = env.get_template("template-eng-cal.tex")
            else:
                jtemplate = env.get_template(self.__main_file_name)

            logoPath = self.__latexTemplatePath / "logo.png"

            # override logo path
            if self.__properties.report_logo is not None:
                logoPath = self.__properties.report_logo

            self.__templated = jtemplate.render(
                place_holder=self.__sepForTex.join(united),
                glossary=self.__opt_makeGlossary,
                logo_path=logoPath.as_posix(),
                project_brief=project_brief,
                module_name=self.__properties.module_name,
                module_version=self.__properties.module_version,
                report_designer=self.__properties.report_designer,
                report_date=self.__properties.report_date,
                report_time=self.__properties.report_time,
                report_token=self.__properties.report_token,
            )

        elif template == ReportTemplateEnum.TEX_KOMA:
            print("... template TEX_KOMA")
            jtemplate = env.get_template("template-koma.tex")
            self.__templated = jtemplate.render(
                glossary=self.__opt_makeGlossary,
                place_holder=self.__sepForTex.join(united),
            )

        else:
            print("ERR: driver unknown !!! quit")
            return False

        return True


class MarkdownReporter:
    """Class that compiles markdown fragments to produce PDF, MD, or HTML.

    This class compiles markdown fragments using weasyprint to produce PDF files,
    or can output raw markdown and HTML.

    Typical usage example:

    ```
     reporter = MarkdownReporter(markdownTemplatePath)
     reporter.linkFragments(ReportTemplateEnum.MD_ENG_CAL, [frag])
     reporter.compileDocument(path=str(working_path), fileName="report")
    ```
    """
    def __init__(self, markdownTemplatePath: Path):
        self.__main_file_name: str = ""
        self.__templated: Union[str, None] = ""
        self.__sep: str = "\n\n"
        self.__opt_makeGlossary: bool = False
        self.__ll: Literal[0, 1, 2, 3] = 3
        self.__markdownTemplatePath: Path = markdownTemplatePath
        self.__properties: ReportProperties = ReportProperties()

    def setProperties(self, prop: ReportProperties) -> None:
        self.__properties = prop

    def compileDocument(
        self,
        path: str = "",
        fileName: str = "report",
        output_format: Literal["md", "html", "pdf"] = "md",
        deleteIfExists: bool = True,
    ) -> None:
        """Build output file from markdown.

        Write the templated content to <path> with name <fileName>.
        Extensions are automatically added at the end of file name.

        Args:
            path (str, optional): where output files are created. Defaults to ''.
            fileName (str, optional): input file name. Defaults to 'report'.
            output_format (Literal["md", "html", "pdf"], optional):
                Output format. Defaults to 'md'.
            deleteIfExists (bool, optional): if True delete file if exists
        """
        import markdown # type: ignore[import-untyped]

        completeFileName = Path(f"{path}/{fileName}")
        completeFileNameMd = Path(f"{completeFileName}.md")
        completeFileNameHtml = Path(f"{completeFileName}.html")

        # Always save markdown file
        with open(file=completeFileNameMd, mode="w", encoding="utf-8") as f:
            f.write(self.__templated)  # type: ignore
        logger.log(
            tp="INF", level=self.__ll, msg=f"Markdown file saved to {completeFileNameMd}"
        )

        if output_format in ("html", "pdf"):
            # Convert markdown to HTML
            md = markdown.Markdown(extensions=["tables", "fenced_code", "md_in_html"])
            html_content = md.convert(self.__templated)  # type: ignore

            # Load CSS
            css_path = self.__markdownTemplatePath / "default.css"
            css_content = ""
            if css_path.exists():
                with open(css_path, encoding="utf-8") as css_file:
                    css_content = css_file.read()

            # MathJax configuration for browser rendering (HTML output)
            mathjax_script = """
    <script>
        MathJax = {
            tex: {
                inlineMath: [['$', '$']],
                displayMath: [['$$', '$$']]
            },
            svg: {
                fontCache: 'global'
            }
        };
    </script>
    <script src="https://cdn.jsdelivr.net/npm/mathjax@3/es5/tex-svg.js" async></script>"""

            # Build full HTML document with MathJax for LaTeX rendering
            html_doc = f"""<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <title>{self.__properties.project_brief}</title>
    <style>
{css_content}
    </style>{mathjax_script}
</head>
<body>
{html_content}
</body>
</html>"""

            if output_format == "html" or output_format == "pdf":
                if deleteIfExists and completeFileNameHtml.exists():
                    completeFileNameHtml.unlink()
                with open(file=completeFileNameHtml, mode="w", encoding="utf-8") as f:
                    f.write(html_doc)
                logger.log(
                    tp="INF", level=self.__ll, msg=f"HTML file saved to {completeFileNameHtml}"
                )

            if output_format == "pdf":
                logger.log(
                    tp="WRN", level=self.__ll,
                    msg="PDF output from markdown is not yet supported. Use HTML output instead."
                )
                print("WRN: PDF output from markdown is not yet supported. Use HTML output instead.")

    def linkFragments(
        self,
        template: ReportTemplateEnum,
        fragments: List[MarkdownFragment] | None = None,
        builder: Union[MarkdownFragmentsBuilder, List[MarkdownFragmentsBuilder]] | None = None,
        glossary: bool = False,
        main_file_name: str = "",
    ) -> bool:
        """Links markdown fragments adding also a main document.

        Args:
            template (ReportTemplateEnum):
             Enum for main document template (MD_ENG_CAL or MD_MAIN).
            fragments (List[MarkdownFragment], Optional):
             List of markdown fragments to put in main document.
            builder (Union[MarkdownFragmentsBuilder, List[MarkdownFragmentsBuilder]], Optional):
             List of fragment builder. This is alternative to fragments
             argument. Default is None.
            glossary (bool, optional): Default is False.
            main_file_name (str, optional): Default is "".

        Returns:
            True on success, False otherwise
        """
        self.__opt_makeGlossary = glossary
        self.__main_file_name = main_file_name
        print("buildMarkdown: start ...")

        if builder is not None:
            if not isinstance(builder, list):
                builders = [builder]
            else:
                builders = builder

            fragments = []
            for b in builders:
                fragments.append(b.buildFragment())

        united: List[str] = []
        assert fragments is not None
        for f in fragments:
            if f.frags() is not None:
                united += f.frags()
            else:
                print("ERR: None fragment")
                print("buildMarkdown: Quit")
                return False

        file_loader = FileSystemLoader(searchpath=self.__markdownTemplatePath)
        env = Environment(
            trim_blocks=True,
            autoescape=False,
            loader=file_loader,
        )

        print("build with markdown ...")

        if (
            template == ReportTemplateEnum.MD_ENG_CAL
            or template == ReportTemplateEnum.MD_MAIN
        ):
            print("... template MD_ENG_CAL")
            if template == ReportTemplateEnum.MD_ENG_CAL:
                jtemplate = env.get_template("template-eng-cal.md")
            else:
                jtemplate = env.get_template(self.__main_file_name)

            logoPath = self.__markdownTemplatePath / "logo.png"

            # override logo path
            if self.__properties.report_logo is not None:
                logoPath = self.__properties.report_logo

            self.__templated = jtemplate.render(
                place_holder=self.__sep.join(united),
                glossary=self.__opt_makeGlossary,
                logo_path=logoPath.as_posix(),
                project_brief=self.__properties.project_brief,
                module_name=self.__properties.module_name,
                module_version=self.__properties.module_version,
                report_designer=self.__properties.report_designer,
                report_date=self.__properties.report_date,
                report_time=self.__properties.report_time,
                report_token=self.__properties.report_token,
                page="1",
            )
        else:
            print("ERR: template unknown !!! quit")
            return False

        return True


class TypstFragment:
    """Class that holds fragments as list of strings for use with TypstReporter.

    This class is a container for complete a typst template.
    Templates use custom Jinja2 delimiters to avoid Typst '#' conflicts:
    block: <%% ... %%>, variable: << ... >>, comment: <#% ... %#>
    """
    def __init__(self, typstTemplatePath: Path = Path()):
        self.__fragment: List[str] = []
        self.__typstTemplatePath: Path = typstTemplatePath
        self.__logoFilePath: Path = Path(self.__typstTemplatePath) / Path("logo.png")

    def set_logo(self, logo_path: Path) -> None:
        self.__logoFilePath = logo_path

    def add_line(self, line: str) -> None:
        self.__fragment.append(line)

    def add_lines(self, lines: List[str]) -> None:
        for ll in lines:
            if ll is not None:
                self.__fragment.append(ll)

    def add_template(self, name: str, placeholders: Dict[str, Any]) -> None:
        """Add a rendered template using custom Jinja2 delimiters for Typst."""
        file_loader = FileSystemLoader(searchpath=self.__typstTemplatePath)
        env = Environment(
            block_start_string="<%%",
            block_end_string="%%>",
            variable_start_string="<<",
            variable_end_string=">>",
            comment_start_string="<#%",
            comment_end_string="%#>",
            trim_blocks=True,
            autoescape=False,
            loader=file_loader,
        )
        jtemplate = env.get_template(name)
        rendered = jtemplate.render(placeholders, logo_path=self.__logoFilePath.as_posix())
        self.__fragment.append(rendered)

    def add(
        self,
        line: Union[str, None] = None,
        lines: Union[List[str], None] = None,
        templateName: Union[str, None] = None,
        templatePlaceholders: Union[Dict[str, Any], None] = None,
    ) -> int:
        if (
            line is not None
            and templateName is None
            and templatePlaceholders is None
            and lines is None
        ):
            self.add_line(line)
        elif (
            line is None
            and lines is None
            and templateName is not None
            and templatePlaceholders is not None
        ):
            self.add_template(templateName, templatePlaceholders)
        elif (
            line is None
            and lines is not None
            and templateName is None
            and templatePlaceholders is None
        ):
            self.add_lines(lines)
        else:
            raise EXAExceptions("0001", "Wrong args", "")
        return len(self.__fragment)

    def frags(self) -> List[str]:
        return self.__fragment


class TypstFragmentsBuilder:
    """Abstract base class for typst fragment builders."""
    def __init__(self) -> None:
        self.__fragment: Union[TypstFragment, None] = None
        self.__fragment_options: Dict[str, Any] = {}

    def _setFragment(self, f: TypstFragment) -> None:
        self.__fragment = f

    def getFragmentOptions(self) -> Dict[str, Any]:
        return self.__fragment_options

    def _setFragmentOptions(self, options: Dict[str, Any]) -> bool:
        self.__fragment_options = options
        return True

    def setFragmentOptions(self, options: Dict[str, Any]) -> bool:
        for name_opt, value_opt in options.items():
            if name_opt not in self.__fragment_options.keys():
                raise ValueError(f"{name_opt} is not a valid option")
        self.__fragment_options = options
        return True

    def buildFragment(self) -> TypstFragment:
        raise NotImplementedError("Need to be implemented")

    def fragment(self) -> Union[TypstFragment, None]:
        return self.__fragment


class TypstReporter:
    """Class that compiles typst fragments to produce PDF or .typ files.

    This class compiles typst fragments using the typst Python package
    to produce PDF files, or can output raw .typ source.

    Typical usage example:

    ```
     reporter = TypstReporter(typstTemplatePath)
     reporter.linkFragments(ReportTemplateEnum.TYP_ENG_CAL, [frag])
     reporter.compileDocument(path=str(working_path), fileName="report")
    ```
    """
    def __init__(self, typstTemplatePath: Path):
        self.__main_file_name: str = ""
        self.__templated: Union[str, None] = ""
        self.__sep: str = "\n\n"
        self.__ll: Literal[0, 1, 2, 3] = 3
        self.__typstTemplatePath: Path = typstTemplatePath
        self.__properties: ReportProperties = ReportProperties()
        self.__logoSourcePath: Path | None = None

    def setProperties(self, prop: ReportProperties) -> None:
        self.__properties = prop

    def compileDocument(
        self,
        path: str = "",
        fileName: str = "report",
        output_format: Literal["typ", "pdf"] = "typ",
        deleteIfExists: bool = True,
    ) -> None:
        """Build output file from typst.

        Write the templated content to <path> with name <fileName>.
        Extensions are automatically added at the end of file name.

        Args:
            path: where output files are created. Defaults to ''.
            fileName: input file name. Defaults to 'report'.
            output_format: Output format ('typ' or 'pdf'). Defaults to 'typ'.
            deleteIfExists: if True delete file if exists
        """
        completeFileName = Path(f"{path}/{fileName}")
        completeFileNameTyp = Path(f"{completeFileName}.typ")
        completeFileNamePdf = Path(f"{completeFileName}.pdf")

        # Copy logo to output directory if available (for .typ portability)
        if self.__logoSourcePath is not None:
            destLogo = Path(path) / self.__logoSourcePath.name
            if destLogo.resolve() != self.__logoSourcePath.resolve():
                shutil.copy2(self.__logoSourcePath, destLogo)

        # Always save typst file
        with open(file=completeFileNameTyp, mode="w", encoding="utf-8") as f:
            f.write(self.__templated)  # type: ignore
        logger.log(
            tp="INF", level=self.__ll, msg=f"Typst file saved to {completeFileNameTyp}"
        )

        if output_format == "pdf":
            import typst  # type: ignore[import-untyped]

            if deleteIfExists and completeFileNamePdf.exists():
                completeFileNamePdf.unlink()

            pdf_bytes = typst.compile(completeFileNameTyp)
            completeFileNamePdf.write_bytes(pdf_bytes)
            logger.log(
                tp="INF", level=self.__ll,
                msg=f"PDF file saved to {completeFileNamePdf}"
            )

    def linkFragments(
        self,
        template: ReportTemplateEnum,
        fragments: List[TypstFragment] | None = None,
        builder: Union[TypstFragmentsBuilder, List[TypstFragmentsBuilder]] | None = None,
        glossary: bool = False,
        main_file_name: str = "",
    ) -> bool:
        """Links typst fragments adding also a main document.

        Args:
            template: Enum for main document template (TYP_ENG_CAL or TYP_MAIN).
            fragments: List of typst fragments to put in main document.
            builder: List of fragment builder. Alternative to fragments. Default is None.
            glossary: Default is False.
            main_file_name: Default is "".

        Returns:
            True on success, False otherwise
        """
        self.__main_file_name = main_file_name
        print("buildTypst: start ...")

        if builder is not None:
            if not isinstance(builder, list):
                builders = [builder]
            else:
                builders = builder

            fragments = []
            for b in builders:
                fragments.append(b.buildFragment())

        united: List[str] = []
        assert fragments is not None
        for f in fragments:
            if f.frags() is not None:
                united += f.frags()
            else:
                print("ERR: None fragment")
                print("buildTypst: Quit")
                return False

        file_loader = FileSystemLoader(searchpath=self.__typstTemplatePath)
        env = Environment(
            block_start_string="<%%",
            block_end_string="%%>",
            variable_start_string="<<",
            variable_end_string=">>",
            comment_start_string="<#%",
            comment_end_string="%#>",
            trim_blocks=True,
            autoescape=False,
            loader=file_loader,
        )

        print("build with typst ...")

        if (
            template == ReportTemplateEnum.TYP_ENG_CAL
            or template == ReportTemplateEnum.TYP_MAIN
        ):
            print("... template TYP_ENG_CAL")
            if template == ReportTemplateEnum.TYP_ENG_CAL:
                jtemplate = env.get_template("template-eng-cal.typ")
            else:
                jtemplate = env.get_template(self.__main_file_name)

            logoPath = self.__typstTemplatePath / "logo.png"

            # override logo path
            if self.__properties.report_logo is not None:
                logoPath = self.__properties.report_logo

            # Resolve to absolute path to make the copy condition robust.
            # Typst resolves images relative to the .typ file, so pass only the filename.
            self.__logoSourcePath = logoPath.resolve() if logoPath.exists() else None
            logoFileName = logoPath.name if self.__logoSourcePath is not None else ""

            def _esc(s: str) -> str:
                """Escape Typst special characters in user strings."""
                return s.replace("\\", "\\\\").replace("@", "\\@").replace("<", "\\<").replace(">", "\\>").replace("#", "\\#")

            self.__templated = jtemplate.render(
                place_holder=self.__sep.join(united),
                glossary=glossary,
                logo_path=logoFileName,
                has_logo=self.__logoSourcePath is not None,
                project_brief=_esc(self.__properties.project_brief),
                module_name=_esc(self.__properties.module_name),
                module_version=_esc(self.__properties.module_version),
                report_designer=_esc(self.__properties.report_designer),
                report_date=self.__properties.report_date,
                report_time=self.__properties.report_time,
                report_token=self.__properties.report_token,
                page="1",
            )
        else:
            print("ERR: template unknown !!! quit")
            return False

        return True



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from pycivil import EXAGeometry as eg
from pycivil.EXAStructural import codes, materials, sections
from pycivil.EXAStructural import plot as ep

# import numpy as np
# from matplotlib.path import Path
# from matplotlib.patches import PathPatch
# import matplotlib.pyplot as plt
# from matplotlib.ticker import FormatStrFormatter

# ******************* CONCRETE SECTION first method

# Setting working code
code_EC2 = codes.Code("EC2")
print(code_EC2)

# Setting concrete material
cls_material = materials.Concrete("EC2_C35/45")
cls_material.setByCode(code_EC2, "C35/45")
print(cls_material)

# Setting concrete shape
rect_shape = eg.ShapeRect(300, 500)
print(rect_shape)

rectangularSection = sections.StructSectionItem(rect_shape, cls_material)
print(rectangularSection)

# Setting steel material
steel_material = materials.ConcreteSteel("EC2_450C")
steel_material.setByCode(code_EC2, "B450C")
print(steel_material)

# Retriving special points
MB = rectangularSection.getShape().getShapePoint("MB")
MT = rectangularSection.getShape().getShapePoint("MT")

# Setting bottom steel shape
area_shape_1 = eg.ShapeArea(196.6)
area_shape_1.setOrigin(MB + eg.Point2d(0, 40))
steel_1 = sections.StructSectionItem(area_shape_1, steel_material)
print(steel_1)

# Setting top steel shape
area_shape_2 = eg.ShapeArea(0)
area_shape_2.setOrigin(MT + eg.Point2d(0, -40))
steel_2 = sections.StructSectionItem(area_shape_2, steel_material)

# TODO: __str__ for ConcreteSection
myfirstsection = sections.ConcreteSection(1, "300x500 EC2")
myfirstsection.setStructConcrItem(rectangularSection)
myfirstsection.setStructSteelItems([steel_1, steel_2])

myfirstsection.setConcreteMaterial(cls_material)
myfirstsection.setSteelMaterial(steel_material)

print(myfirstsection)

NxMz, Fields, NxMzBoundingRect = myfirstsection.build2dInteractionCompleteDomain()

strDimension = (
    "FERRAILLAGE 1D - CS21 - 300x500 - Ai=196.6 As=0 - Nx = +388KN - Mz = +121.00KN"
)
strMaterial = "concrete: B450C steel: C35/45 gammas: 1.15 gammac: 1.5"

# tensionPoints = [[+0,+58.49]]
tensionPoints = [[388, (123.00 + 119) / 2]]

ep.interactionDomainBasePlot2d(
    NxMz,
    Fields,
    xLabel="Nx [KN]",
    yLabel="Mz [KN*m]",
    titleAddStr=strDimension + "\n" + strMaterial,
    tensionPoints=tensionPoints,
    export="test_benchmark_CS21_no2.png",
    dpi=300,
)



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

import json
from pathlib import Path

from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import ModelInputRcSecCheck
from pycivil.EXAUtils.markdownReportMakers import ConcreteSectionMdFB
from pycivil.EXAUtils.report import (
    MarkdownReporter,
    ReportTemplateEnum,
    getMarkdownTemplatesPath,
)


def test_report_markdown_section_01(tmp_path: Path):
    """Test RC section markdown fragment generation with first test data."""
    with open(Path(__file__).parent / "test_report_latex_sections_01.json") as jsonFile:
        jsonObject = json.load(jsonFile)
        jsonFile.close()

    iData = ModelInputRcSecCheck(**jsonObject)

    print(json.dumps(iData.model_dump(), indent=3))

    section = iData.buildRcsRectangular()
    if section is not None:
        markdownTemplatePath = getMarkdownTemplatesPath()
        f = ConcreteSectionMdFB(markdownTemplatePath, rcs=section).buildFragment()
        print(f.frags())
        reporter = MarkdownReporter(markdownTemplatePath)
        reporter.linkFragments(
            template=ReportTemplateEnum.MD_ENG_CAL, fragments=[f]
        )
        reporter.compileDocument(path=str(tmp_path), output_format="md")
        print(f"Temporary path for test is {tmp_path}")
        assert tmp_path.joinpath("report.md").exists()

        # Verify content has section data
        content = tmp_path.joinpath("report.md").read_text(encoding="utf-8")
        assert "400" in content  # width or height


def test_report_markdown_section_02(tmp_path: Path):
    """Test RC section markdown fragment generation with second test data."""
    with open(Path(__file__).parent / "test_report_latex_sections_02.json") as jsonFile:
        jsonObject = json.load(jsonFile)
        jsonFile.close()

    iData = ModelInputRcSecCheck(**jsonObject)

    print(json.dumps(iData.model_dump(), indent=3))

    section = iData.buildRcsRectangular()
    if section is not None:
        markdownTemplatePath = getMarkdownTemplatesPath()
        f = ConcreteSectionMdFB(markdownTemplatePath, rcs=section).buildFragment()
        print(f.frags())
        reporter = MarkdownReporter(markdownTemplatePath)
        reporter.linkFragments(
            template=ReportTemplateEnum.MD_ENG_CAL, fragments=[f]
        )
        reporter.compileDocument(path=str(tmp_path), output_format="md")
        print(f"Temporary path for test is {tmp_path}")
        assert tmp_path.joinpath("report.md").exists()


def test_report_markdown_section_with_html(tmp_path: Path):
    """Test RC section with HTML output."""
    with open(Path(__file__).parent / "test_report_latex_sections_01.json") as jsonFile:
        jsonObject = json.load(jsonFile)
        jsonFile.close()

    iData = ModelInputRcSecCheck(**jsonObject)

    section = iData.buildRcsRectangular()
    if section is not None:
        markdownTemplatePath = getMarkdownTemplatesPath()
        f = ConcreteSectionMdFB(markdownTemplatePath, rcs=section).buildFragment()
        reporter = MarkdownReporter(markdownTemplatePath)
        reporter.linkFragments(
            template=ReportTemplateEnum.MD_ENG_CAL, fragments=[f]
        )
        reporter.compileDocument(path=str(tmp_path), output_format="html")
        print(f"Temporary path for test is {tmp_path}")
        assert tmp_path.joinpath("report.md").exists()
        assert tmp_path.joinpath("report.html").exists()



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

"""Geometric primitives for 2D and 3D structural analysis.

Provides points, vectors, nodes, segments, edges, and polylines in 2D and 3D
space, along with applied-vector and force-couple representations used
throughout the PyCivil structural pipeline.

Typical units follow the PyCivil convention: millimetres (mm) for lengths,
Newtons (N) for forces, and MPa for stresses.
"""
from __future__ import annotations

import math
from copy import deepcopy
from typing import List, Optional, Tuple, Union

from pydantic import BaseModel


# Defines points or free vector in 2D space
class Point2d:
    """A point (or free vector) in 2D Cartesian space.

    Coordinates are stored as private attributes and exposed via properties.
    The class supports basic arithmetic (``+``, ``-``, scalar ``*``), equality
    testing, and common geometric operations such as rotation, translation, and
    signed-area calculation.

    Attributes:
        x: X coordinate.
        y: Y coordinate.
    """

    def __init__(
        self,
        x: Union[float, int] = 0.0,
        y: Union[float, int] = 0.0,
        make_null: bool = False,
        coords: Union[Tuple[Union[float, int], Union[float, int]], None] = None,
    ):
        """Initialise a Point2d.

        Args:
            x: X coordinate.  Ignored when *coords* is provided.
            y: Y coordinate.  Ignored when *coords* is provided.
            make_null: When ``True`` the point is flagged as a *null* sentinel
                (e.g. to signal a missing intersection result).
            coords: Optional ``(x, y)`` tuple.  Takes precedence over the
                individual *x* / *y* arguments when supplied.

        Raises:
            ValueError: If *x* is not a ``float`` or ``int``.
        """
        if isinstance(x, (float, int)):
            self.__x = x
            self.__y = y
            self.__isNull: bool = make_null
        else:
            raise ValueError("Point2d init only with float or int !!!")

        if coords is not None:
            self.__x = coords[0]
            self.__y = coords[1]

    def __str__(self):
        return f"Point2d: ({self.__x:.2e},{self.__y:.2e})"

    def __add__(self, other: Point2d) -> Point2d:
        return Point2d(self.__x + other.x, self.__y + other.y)

    def __sub__(self, other: Point2d) -> Point2d:
        return Point2d(self.__x - other.x, self.__y - other.y)

    def __mul__(self, other: Union[float, int]) -> Point2d:
        return Point2d(self.__x * other, self.__y * other)

    def __rmul__(self, other: float) -> Point2d:
        return Point2d(self.__x * other, self.__y * other)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Point2d):
            return NotImplemented
        if self.__x == other.x and self.__y == other.y:
            return True
        else:
            return False

    def __repr__(self):
        return str(self)

    @property
    def x(self):
        """float: X coordinate."""
        return self.__x

    @x.setter
    def x(self, val: float) -> None:
        self.__x = val

    @property
    def y(self):
        """float: Y coordinate."""
        return self.__y

    @y.setter
    def y(self, val: float) -> None:
        self.__y = val

    def isNull(self) -> bool:
        """Return ``True`` if this point was created as a null sentinel.

        Returns:
            bool: ``True`` when the point carries no geometric meaning.
        """
        return self.__isNull

    def isEqualTo(self, other, rox, roy, prec):
        """Test approximate equality with reference scales and a tolerance.

        The comparison is performed in normalised coordinates: each difference
        is divided by the corresponding reference length before checking
        against *prec*.

        Args:
            other: Another ``Point2d`` to compare against.
            rox: Reference length along the X axis used for normalisation.
            roy: Reference length along the Y axis used for normalisation.
            prec: Tolerance threshold in normalised units.

        Returns:
            bool: ``True`` if both normalised differences are smaller than
                *prec*.
        """
        dx = math.sqrt(math.pow(self.__x - other.x, 2) / math.pow(rox, 2))
        dy = math.sqrt(math.pow(self.__y - other.y, 2) / math.pow(roy, 2))
        if dx < prec and dy < prec:
            return True
        else:
            return False

    def distance(self, other) -> float:
        """Return the Euclidean distance to another point.

        Args:
            other: Target ``Point2d``.

        Returns:
            float: Distance between the two points.
        """
        return math.sqrt(
            math.pow(self.__x - other.x, 2) + math.pow(self.__y - other.y, 2)
        )

    def midpoint(self, other) -> Point2d:
        """Return the midpoint between this point and *other*.

        Args:
            other: The second ``Point2d``.

        Returns:
            Point2d: The midpoint.
        """
        return Point2d((self.__x + other.x) / 2, (self.__y + other.y) / 2)

    def cross(self, p1, p2) -> float:
        """Compute the signed cross product (z-component) relative to this point.

        Geometrically this equals twice the signed area of the triangle
        ``(self, p1, p2)``.  A positive result means *p2* is to the left of
        the directed edge ``self → p1``.

        Args:
            p1: First ``Point2d``.
            p2: Second ``Point2d``.

        Returns:
            float: Signed cross product value.
        """
        x1 = p1.x - self.__x
        y1 = p1.y - self.__y

        x2 = p2.x - self.__x
        y2 = p2.y - self.__y

        return x1 * y2 - x2 * y1

    def affine(self, m11: float, m12: float, m21: float, m22: float, dx: float, dy: float) -> Point2d:
        """Apply a general 2-D affine transformation in-place.

        The transformation is defined as::

            x' = m11*x + m12*y + dx
            y' = m21*x + m22*y + dy

        Args:
            m11: Element (1,1) of the linear part.
            m12: Element (1,2) of the linear part.
            m21: Element (2,1) of the linear part.
            m22: Element (2,2) of the linear part.
            dx: Translation in X.
            dy: Translation in Y.

        Returns:
            Point2d: *self* after the transformation (allows chaining).
        """
        x_old = self.__x
        y_old = self.__y
        self.__x = m11 * x_old + m12 * y_old + dx
        self.__y = m21 * x_old + m22 * y_old + dy
        return self

    def rotate(self, angle: float, center: Point2d) -> Point2d:
        """Rotate the point in-place around a given centre.

        Args:
            angle: Rotation angle in **degrees** (positive = counter-clockwise).
            center: Centre of rotation.

        Returns:
            Point2d: *self* after the rotation (allows chaining).
        """
        alpha = math.radians(angle)
        self.affine(1, 0, 0, 1, -center.x, -center.y)
        self.affine(math.cos(alpha), -math.sin(alpha), math.sin(alpha), math.cos(alpha), 0, 0)
        self.affine(1, 0, 0, 1, +center.x, +center.y)
        return self

    def translate(self, dx: float, dy: float) -> Point2d:
        """Translate the point in-place.

        Args:
            dx: Displacement along the X axis.
            dy: Displacement along the Y axis.

        Returns:
            Point2d: *self* after the translation (allows chaining).
        """
        self.affine(1, 0, 0, 1, dx, dy)
        return self

    @staticmethod
    def areaFromTria(p0: Point2d, p1: Point2d, p2: Point2d) -> float:
        """Return the area of a triangle defined by three points.

        Uses the standard cross-product (shoelace) formula.

        Args:
            p0: First vertex.
            p1: Second vertex.
            p2: Third vertex.

        Returns:
            float: Non-negative area of the triangle.
        """
        assert isinstance(p0.x, (float, int))
        assert isinstance(p0.y, (float, int))
        assert isinstance(p1.x, (float, int))
        assert isinstance(p1.y, (float, int))
        assert isinstance(p2.x, (float, int))
        assert isinstance(p2.y, (float, int))
        return (
            1
            / 2
            * abs(
                p0.x * p1.y * 1
                + p0.y * 1 * p2.x
                + 1 * p1.x * p2.y
                - 1 * p1.y * p2.x
                - p0.y * p1.x * 1
                - p0.x * 1 * p2.y
            )
        )


def boundingBox(p1, p2):
    """Return the axis-aligned bounding box of two 2-D points.

    Args:
        p1: First ``Point2d``.
        p2: Second ``Point2d``.

    Returns:
        list[float]: ``[min_x, max_x, min_y, max_y]``.

    Raises:
        Exception: If either argument is not a ``Point2d``.
    """
    if not isinstance(p1, Point2d) or not isinstance(p2, Point2d):
        raise Exception("Bounding only for two Point2d !!!")

    assert isinstance(p1.x, (float, int))
    assert isinstance(p1.y, (float, int))
    assert isinstance(p2.x, (float, int))
    assert isinstance(p2.y, (float, int))

    min_x = min(p1.x, p2.x)
    max_x = max(p1.x, p2.x)
    min_y = min(p1.y, p2.y)
    max_y = max(p1.y, p2.y)

    return [min_x, max_x, min_y, max_y]


class Node2d(Point2d):
    """A 2-D mesh node extending ``Point2d`` with an integer identifier.

    Attributes:
        idn: Integer node identifier (default ``-1`` for unassigned).
    """

    def __init__(self, x=0.0, y=0.0, idn=-1):
        """Initialise a Node2d.

        Args:
            x: X coordinate.
            y: Y coordinate.
            idn: Node identifier.
        """
        Point2d.__init__(self, x, y)
        self.idn = idn

    def __str__(self):
        return f"Node2D: idn = {self.idn:.0f}, x = {self.__x:.3e}, y = {self.__y:.3e}"


# Defines points or free vector in 3D space
class Point3d:
    """A point (or free vector) in 3D Cartesian space.

    Attributes:
        x: X coordinate.
        y: Y coordinate.
        z: Z coordinate.
    """

    def __init__(self, x=0.0, y=0.0, z=0.0):
        """Initialise a Point3d.

        Args:
            x: X coordinate.
            y: Y coordinate.
            z: Z coordinate.
        """
        self.__x = x
        self.__y = y
        self.__z = z

    def __str__(self) -> str:
        return f"Point3D: x = {self.__x:.3e}, y = {self.__y:.3e}, z = {self.__z:.3e}"

    def __add__(self, other: object) -> Point3d:
        if not isinstance(other, Point3d):
            return NotImplemented

        return Point3d(self.__x + other.x, self.__y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Point3d(self.__x - other.x, self.__y - other.y, self.z - other.z)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Point3d):
            return NotImplemented

        if self.__x == other.x and self.__y == other.y and self.__z == other.z:
            return True
        else:
            return False

    @property
    def x(self):
        """float: X coordinate."""
        return self.__x

    @x.setter
    def x(self, val: float) -> None:
        self.__x = val

    @property
    def y(self):
        """float: Y coordinate."""
        return self.__y

    @y.setter
    def y(self, val: float) -> None:
        self.__y = val

    @property
    def z(self):
        """float: Z coordinate."""
        return self.__z

    @z.setter
    def z(self, val: float) -> None:
        self.__z = val

    def tranlate(self, dx=0.0, dy=0.0, dz=0.0):
        """Translate the point in-place.

        Args:
            dx: Displacement along the X axis.
            dy: Displacement along the Y axis.
            dz: Displacement along the Z axis.
        """
        self.__x = self.__x + dx
        self.__y = self.__y + dy
        self.z = self.z + dz

    def distanceFrom(self, p: Point3d) -> float:
        """Return the Euclidean distance to another 3-D point.

        Args:
            p: Target ``Point3d``.

        Returns:
            float: Distance between the two points.

        Raises:
            Exception: If *p* is not a ``Point3d`` instance.
        """
        if not isinstance(p, Point3d):
            raise Exception("Arg must be a Point3d instance !!!")

        return math.dist([self.__x, self.__y, self.z], [p.x, p.y, p.z])


def areaFromTria3D(p0: Point3d, p1: Point3d, p2: Point3d) -> float:
    """Return the area of a 3-D triangle defined by three points.

    Uses the cross-product formula:
    ``area = 0.5 * |v × w|``
    where *v* = ``p0→p1`` and *w* = ``p1→p2``.

    Args:
        p0: First vertex.
        p1: Second vertex.
        p2: Third vertex.

    Returns:
        float: Non-negative area of the triangle.
    """
    v = Vector3d(p0, p1)
    w = Vector3d(p1, p2)
    return 1 / 2 * math.sqrt(abs(v.scalar(v) * w.scalar(w) - v.scalar(w) * v.scalar(w)))


class Node3d(Point3d):
    """A 3-D mesh node extending ``Point3d`` with an integer identifier.

    Attributes:
        idn: Integer node identifier (default ``-1`` for unassigned).
    """

    def __init__(self, x=0.0, y=0.0, z=0.0, idn=-1):
        """Initialise a Node3d.

        Args:
            x: X coordinate.
            y: Y coordinate.
            z: Z coordinate.
            idn: Node identifier.
        """
        Point3d.__init__(self, x, y, z)
        self.idn = idn

    def __str__(self) -> str:
        return "Node3D: idn = {:.0f}, x = {:.3e}, y = {:.3e}, z = {:.3e}".format(
            self.idn,
            self.x,
            self.y,
            self.z,
        )

    def __add__(self, other: object) -> Node3d:
        if not isinstance(other, Node3d):
            return NotImplemented

        return Node3d(self.__x + other.x, self.__y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Node3d(self.__x - other.x, self.__y - other.y, self.z - other.z)


class Seg2d:
    """A directed line segment in 2-D space defined by two ``Point2d`` endpoints.

    Provides intersection, containment, and factor utilities needed for
    polygon clipping and section outline processing.

    Attributes:
        p0: Start point of the segment.
        p1: End point of the segment.
    """

    def __init__(self, p0: Point2d, p1: Point2d):
        """Initialise a Seg2d from two endpoints.

        Args:
            p0: Start point.
            p1: End point.
        """
        self.__p0 = p0
        self.__p1 = p1

    @property
    def p0(self):
        """Point2d: Start point."""
        return self.__p0

    @p0.setter
    def p0(self, val: Point2d) -> None:
        self.__p0 = val

    @property
    def p1(self):
        """Point2d: End point."""
        return self.__p1

    @p1.setter
    def p1(self, val: Point2d) -> None:
        self.__p1 = val

    def len(self) -> float:
        """Return the length of the segment.

        Returns:
            float: Euclidean distance between *p0* and *p1*.
        """
        return self.__p0.distance(self.__p1)

    # Find the intersection point by Seg2D in input
    def intersect(self, other: Seg2d) -> Point2d:
        """Return the intersection point of the two infinite lines.

        The lines are the infinite extensions of *self* and *other*.  If the
        lines are parallel (determinant == 0) a null ``Point2d`` is returned.

        Args:
            other: The second segment (treated as an infinite line).

        Returns:
            Point2d: Intersection point, or a null ``Point2d`` if parallel.
        """
        a0 = self.__p1.x - self.__p0.x
        b0 = other.p0.x - other.p1.x
        c0 = other.p0.x - self.__p0.x

        a1 = self.__p1.y - self.__p0.y
        b1 = other.p0.y - other.p1.y
        c1 = other.p0.y - self.__p0.y

        detA = a0 * b1 - a1 * b0
        if detA == 0:
            return Point2d(make_null=True)

        detA_alpha = c0 * b1 - c1 * b0
        alpha = detA_alpha / detA
        return self.__p0 + alpha * (self.__p1 - self.__p0)

    def boundingHasPoint(self, p: Point2d, toll: float = 1e-15) -> bool:
        """Test whether *p* lies within the axis-aligned bounding box of this segment.

        Args:
            p: The point to test.
            toll: Absolute tolerance added to the bounding box extents.

        Returns:
            bool: ``True`` if *p* is inside (or on) the expanded bounding box.
        """
        bb = boundingBox(self.__p0, self.__p1)
        if (
            p.x < bb[0] - toll
            or p.x > bb[1] + toll
            or p.y < bb[2] - toll
            or p.y > bb[3] + toll
        ):
            return False
        else:
            return True

    # Calculate intersection point from extension of other and self
    def intersectAndContaints(self, other: Seg2d) -> Tuple[bool, Point2d]:
        """Intersect two segments and report whether the intersection lies on *self*.

        Uses a length-ratio test: the intersection belongs to *self* when the
        sum of its distances to *p0* and *p1* equals the segment length
        (within numerical tolerance).

        Args:
            other: The second segment.

        Returns:
            Tuple[bool, Point2d]: A ``(contained, point)`` pair where
                *contained* is ``True`` if the intersection point lies on
                *self*, and *point* is the intersection ``Point2d``.
        """
        pint = self.intersect(other)
        m1 = (pint.distance(self.__p0) + pint.distance(self.__p1))/ self.__p0.distance(self.__p1)
        # m2 = (pint.distance(other.p0) + pint.distance(other.p1)) / other.len()
        pint.isNull()
        if abs(m1-1) < 0.0000000001:
            return True, pint
        else:
            return False, pint

        # if not pint.isNull() and self.boundingHasPoint(pint, toll=self.len() / 100000):
        #     return True, pint
        # else:
        #     return False, pint

    # If p is a Point2d calculs factor witch factor * Seg2d = p
    def extensionFactor(self, p) -> float:
        """Compute the scalar factor such that ``factor * self == p``.

        Useful for parametric position queries along the infinite line through
        *self*.  The Y component is preferred when the X component of the
        segment direction is zero.

        Args:
            p: Target ``Point2d``.

        Returns:
            float: Extension factor along the segment direction.
        """
        vx = self.__p0.x - self.__p1.x
        vy = self.__p0.y - self.__p1.y

        evx = p.x - self.__p1.x
        evy = p.y - self.__p1.y

        f = 1
        if vx != 0.0:
            f = evx / vx
        if vy != 0.0:
            f = evy / vy
        return f

    def __str__(self):
        dispstr = "Point2d i: x = {:.3e}, y = {:.3e} \n".format(
            self.__p0.x,
            self.__p0.y,
        )
        dispstr = dispstr + "Point2d j: x = {:.3e}, y = {:.3e} \n".format(
            self.__p1.x,
            self.__p1.y,
        )
        return dispstr


class Edge2d:
    """A directed edge in a 2-D mesh connecting two ``Node2d`` instances.

    Unlike ``Seg2d``, an ``Edge2d`` carries node identifiers and is intended
    for use in mesh connectivity structures.
    """

    def __init__(self, *args):
        """Initialise an Edge2d from exactly two ``Node2d`` arguments.

        Args:
            *args: Exactly two ``Node2d`` instances (node_i, node_j).

        Raises:
            TypeError: If the arguments are not two ``Node2d`` instances.
        """
        if len(args) == 2:
            if isinstance(args[0], Node2d) & isinstance(args[1], Node2d) is True:
                self.__node_i = args[0]
                self.__node_j = args[1]
            else:
                raise TypeError("Args must be [Node2D] !!!")
        else:
            raise TypeError("Only two argument !!!")

    def nodeI(self) -> Node2d:
        """Return the start node of the edge.

        Returns:
            Node2d: Start node.
        """
        return self.__node_i

    def nodeJ(self) -> Node2d:
        """Return the end node of the edge.

        Returns:
            Node2d: End node.
        """
        return self.__node_j

    def __str__(self):
        dispstr = "Node2D i: idn = {:.0f}, x = {:.3e}, y = {:.3e} \n".format(
            self.__node_i.idn,
            self.__node_i.x,
            self.__node_i.y,
        )
        dispstr = dispstr + "Node2D j: idn = {:.0f}, x = {:.3e}, y = {:.3e} \n".format(
            self.__node_j.idn,
            self.__node_j.x,
            self.__node_j.y,
        )
        return dispstr


class Edge3d:
    """A directed edge in a 3-D mesh connecting two ``Node3d`` instances.

    Supports three construction signatures:

    * No arguments — creates an edge with two default (origin) nodes.
    * Two ``Node3d`` instances — ``Edge3d(node_i, node_j)``.
    * Eight numeric + id values — ``Edge3d(xi, yi, zi, idni, xj, yj, zj, idnj)``.
    """

    def __init__(self, *args):
        """Initialise an Edge3d.

        Args:
            *args: One of the following signatures:

                * ``()`` — default construction, both nodes at the origin.
                * ``(Node3d, Node3d)`` — explicit node pair.
                * ``(float, float, float, int, float, float, float, int)`` —
                  coordinates and identifiers for node I followed by node J.

        Raises:
            Exception: On invalid argument types or counts.
        """
        if len(args) == 2:
            if isinstance(args[0], Node3d) & isinstance(args[1], Node3d) is True:
                self.__node_i = args[0]
                self.__node_j = args[1]
            else:
                raise Exception("Args must be [Node3D] !!!")
        elif len(args) == 0:
            self.__node_i = Node3d()
            self.__node_j = Node3d()
        elif len(args) == 8:
            for i in range(2):
                if not isinstance(args[0 + 4 * i], float):
                    raise Exception("Arg n. %e must be float !!!" % (0 + i))
                if not isinstance(args[1 + 4 * i], float):
                    raise Exception("Arg n. %e must be float !!!" % (1 + i))
                if not isinstance(args[2 + 4 * i], float):
                    raise Exception("Arg n. %e must be float !!!" % (2 + i))
                if not isinstance(args[3 + 4 * i], int):
                    raise Exception("Arg n. %e must be int !!!" % (3 + i))
            self.__node_i = Node3d(args[0], args[1], args[2], args[3])
            self.__node_j = Node3d(args[4], args[5], args[6], args[7])
        else:
            raise Exception("Wrong arguments !!!")

    def nodeI(self) -> Node3d:
        """Return the start node of the edge.

        Returns:
            Node3d: Start node.
        """
        return self.__node_i

    def nodeJ(self) -> Node3d:
        """Return the end node of the edge.

        Returns:
            Node3d: End node.
        """
        return self.__node_j

    def setNodeI(self, *args):
        """Set the start node.

        Args:
            *args: Exactly one ``Node3d`` instance.

        Raises:
            Exception: If the argument is not a single ``Node3d``.
        """
        if len(args) == 1:
            if isinstance(args[0], Node3d):
                self.__node_i = args[0]
            else:
                raise Exception("Arg must be [Node3D] !!!")
        else:
            raise Exception("Only one argument !!!")

    def setNodeJ(self, *args):
        """Set the end node.

        Args:
            *args: Exactly one ``Node3d`` instance.

        Raises:
            Exception: If the argument is not a single ``Node3d``.
        """
        if len(args) == 1:
            if isinstance(args[0], Node3d):
                self.__node_j = args[0]
            else:
                raise Exception("Arg must be [Node3D] !!!")
        else:
            raise Exception("Only one argument !!!")

    def __str__(self):
        dispstr = (
            "Node3D i: idn = {:.0f}, x = {:.3e}, y = {:.3e}, z = {:.3e} \n".format(
                self.__node_i.idn,
                self.__node_i.x,
                self.__node_i.y,
                self.__node_i.z,
            )
        )
        dispstr = (
            dispstr
            + "Node3D j: idn = {:.0f}, x = {:.3e}, y = {:.3e}, z = {:.3e} \n".format(
                self.__node_j.idn,
                self.__node_j.x,
                self.__node_j.y,
                self.__node_j.z,
            )
        )
        return dispstr

    def lenght(self) -> float:
        """Return the length of the edge.

        Returns:
            float: Euclidean distance between node I and node J.
        """
        return math.sqrt(
            math.pow(self.__node_j.x - self.__node_i.x, 2)
            + math.pow(self.__node_j.y - self.__node_i.y, 2)
            + math.pow(self.__node_j.z - self.__node_i.z, 2)
        )


class Polyline2d:
    """List of Nodes2D

    An ordered sequence of ``Node2d`` vertices representing an open or closed
    2-D polyline (polygon boundary).
    """

    def __init__(self, *args):
        """Initialise a Polyline2d from a list of nodes.

        Args:
            *args: Exactly one positional argument — a ``list`` of ``Node2d``
                instances.

        Raises:
            TypeError: If the argument count is wrong, the argument is not a
                list, or any list element is not a ``Node2d``.
        """
        if len(args) == 0:
            raise TypeError("Only one argument !!!")
        elif len(args) == 1:
            lst = args[0]
            if isinstance(lst, list):
                for i in lst:
                    if not isinstance(i, Node2d):
                        raise TypeError("List must be [Node2D] formed !!!")
                self.__node = lst
            else:
                raise TypeError("First arg must be type List !!!")
        elif len(args) > 1:
            raise TypeError("Only one argument !!!")

    def __str__(self):
        dispstr = "Polyline2d: size is %.0f \n" % len(self.__node)
        for n in self.__node:
            dispstr = (
                dispstr
                + "Node2D: idn = {:.0f}, x = {:.3e}, y = {:.3e} \n".format(
                    n.idn,
                    n.x,
                    n.y,
                )
            )
        return dispstr

    def getNodes(self) -> list:
        """Return the underlying list of nodes.

        Returns:
            list[Node2d]: The node list (not a copy).
        """
        return self.__node

    def size(self) -> int:
        """Return the number of nodes in the polyline.

        Returns:
            int: Node count.
        """
        return len(self.__node)

    def setClosed(self) -> None:
        """Close the polyline by appending the first node at the end."""
        self.__node.append(self.__node[0])
        return

    def isClosed(self) -> bool:
        """Return ``True`` if the first and last nodes are the same.

        Returns:
            bool: ``True`` when the polyline is closed.
        """
        sz = len(self.__node)
        if self.__node[0] == self.__node[sz - 1]:
            return True
        else:
            return False

    def translate(
        self, pStart: Union[Point2d, None], pEnd: Union[Point2d, None]
    ) -> None:
        """Translate all nodes by the vector ``pEnd - pStart``.

        Args:
            pStart: Origin of the translation vector.  Defaults to the
                coordinate origin when ``None``.
            pEnd: Tip of the translation vector.  Defaults to the coordinate
                origin when ``None``.
        """
        if pStart is None:
            pStart = Point2d()
        if pEnd is None:
            pEnd = Point2d()

        v = pEnd - pStart
        for node in self.__node:
            node.translate(v.x, v.y)

    def vertexNb(self) -> int:
        """Return the number of vertices in the polyline.

        Returns:
            int: Vertex count (alias for :meth:`size`).
        """
        return len(self.__node)

    def vertexAt(self, i):
        """Return the number of sub-vertices at position *i*.

        Args:
            i: Index into the node list.

        Returns:
            int: Length of the node at position *i* (typically always 1 for
                plain ``Node2d`` objects).
        """
        return len(self.__node[i])


class Polyline3d:
    """List of Nodes3D

    An ordered sequence of ``Node3d`` vertices representing an open or closed
    3-D polyline.
    """

    def __init__(self, *args):
        """Initialise a Polyline3d from a list of nodes.

        Args:
            *args: Exactly one positional argument — a ``list`` of ``Node3d``
                instances.

        Raises:
            Exception: If the argument count is wrong, the argument is not a
                list, or any list element is not a ``Node3d``.
        """
        if len(args) == 0:
            raise Exception("Only one argument !!!")
        elif len(args) == 1:
            lst = args[0]
            if isinstance(lst, list):
                for i in lst:
                    if not isinstance(i, Node3d):
                        raise Exception("List must be [Node3D] formed !!!")
                self.__node = lst
            else:
                raise Exception("First arg must be type List !!!")
        elif len(args) > 1:
            raise Exception("Only one argument !!!")

    def __str__(self):
        dispstr = "Polyline3d: size is %.0f \n" % len(self.__node)
        for n in self.__node:
            dispstr = (
                dispstr
                + "Node3D: idn = {:.0f}, x = {:.3e}, y = {:.3e}, z = {:.3e}\n".format(
                    n.idn,
                    n.x,
                    n.y,
                    n.z,
                )
            )
        return dispstr

    def setClosed(self) -> None:
        """Close the polyline by appending the first node at the end."""
        self.__node.append(self.__node[0])
        return

    def isClosed(self) -> bool:
        """Return ``True`` if the first and last nodes are the same.

        Returns:
            bool: ``True`` when the polyline is closed.
        """
        sz = len(self.__node)
        if self.__node[0] == self.__node[sz - 1]:
            return True
        else:
            return False

    def vertexNb(self) -> int:
        """Return the number of vertices in the polyline.

        Returns:
            int: Vertex count.
        """
        return len(self.__node)

    def vertexAt(self, i):
        """Return the number of sub-vertices at position *i*.

        Args:
            i: Index into the node list.

        Returns:
            int: Length of the node at position *i*.
        """
        return len(self.__node[i])


# Defines points or free vector in 3D space
class Vector2d:
    """A free vector in 2-D space.

    Can be constructed either from two ``Point2d`` endpoints or by providing
    the components directly.

    Attributes:
        vx: X component.
        vy: Y component.
    """

    def __init__(
        self,
        p0: Point2d = Point2d(make_null=True),
        p1: Point2d = Point2d(make_null=True),
        vx: float = 0.0,
        vy: float = 0.0,
    ):
        """Initialise a Vector2d.

        When *p0* and *p1* are non-null ``Point2d`` instances the vector is
        set to ``p1 - p0``.  Otherwise *vx* / *vy* are used directly.

        Args:
            p0: Start point (optional).
            p1: End point (optional).
            vx: X component used when *p0* / *p1* are null.
            vy: Y component used when *p0* / *p1* are null.

        Raises:
            TypeError: If *p0* or *p1* are not ``Point2d`` instances.
        """
        if isinstance(p0, Point2d) and isinstance(p1, Point2d):
            if not p0.isNull() and not p1.isNull():
                self.vx = p1.x - p0.x
                self.vy = p1.y - p0.y
                return
        else:
            raise TypeError("p0 and p1 must be Point2d instances")

        self.vx = vx
        self.vy = vy

    def __str__(self):
        return f"Vector2D: vx = {self.vx:.3e}, vy = {self.vy:.3e}"

    def __add__(self, other: Vector2d) -> Vector2d:
        return Vector2d(vx=self.vx + other.vx, vy=self.vy + other.vy)

    def __sub__(self, other: Vector2d) -> Vector2d:
        if not isinstance(other, Vector2d):
            return NotImplemented
        return Vector2d(vx=self.vx - other.vx, vy=self.vy - other.vy)

    # def __mul__(self, scale):
    #    return Vector3d(self.vx*scale,self.vy*scale,self.z*scale)

    def __rmul__(self, scale):
        return Vector2d(vx=self.vx * scale, vy=self.vy * scale)

    #       _    _    _            _         _         _
    #     | x    y    z  |   y1*z2*x + z1*x2*y + x1*y2*z +
    #  det| x1   y1   z1 | =       _         _         _
    #     | x2   y2   z2 |  -y2*z1*x - x1*z2*y - x2*y1*z
    #
    def cross(self, other: Vector2d) -> float:
        """Compute the scalar cross product (z-component of 3-D cross).

        Args:
            other: The second ``Vector2d``.

        Returns:
            float: ``vx * other.vy - other.vx * vy``.
        """
        x1 = self.vx
        y1 = self.vy

        x2 = other.vx
        y2 = other.vy

        return x1 * y2 - x2 * y1

    def norm(self) -> float:
        """Return the Euclidean norm of the vector.

        Returns:
            float: ``sqrt(vx² + vy²)``.
        """
        return math.sqrt(math.pow(self.vx, 2) + math.pow(self.vy, 2))

    def norm_roxroy(self, rox=1.0, roy=1.0) -> float:
        """Return the norm scaled by reference lengths along each axis.

        Args:
            rox: Reference length for the X component.
            roy: Reference length for the Y component.

        Returns:
            float: ``sqrt((vx/rox)² + (vy/roy)²)``.
        """
        return math.sqrt(math.pow(self.vx / rox, 2) + math.pow(self.vy / roy, 2))

    def normalize(self) -> Vector2d:
        """Normalise the vector in-place to unit length.

        Returns:
            Vector2d: *self* after normalisation (allows chaining).

        Raises:
            Exception: If the vector is the zero vector.
        """
        n = self.norm()
        if n != 0:
            self.vx = self.vx / n
            self.vy = self.vy / n
        else:
            raise Exception("Vector 2D null !!!")
        return self

    def rotate(self, angle: float) -> Vector2d:
        """Rotate the vector in-place around the origin.

        Args:
            angle: Rotation angle in **degrees** (positive = counter-clockwise).

        Returns:
            Vector2d: *self* after rotation (allows chaining).
        """
        hold_vx = self.vx
        hold_vy = self.vy
        alpha = math.radians(angle)
        self.vx = (math.cos(alpha)) * hold_vx + (-math.sin(alpha)) * hold_vy
        self.vy = (math.cos(alpha)) * hold_vy + (math.sin(alpha)) * hold_vx

        return self


# Defines points or free vector in 3D space
class Vector3d:
    """A free vector in 3-D space.

    Supports multiple construction signatures:

    * No arguments — zero vector.
    * Three scalars — ``Vector3d(vx, vy, vz)``.
    * Two ``Point3d`` instances — ``Vector3d(p0, p1)`` → ``p1 - p0``.
    * One 3-tuple of scalars — ``Vector3d((vx, vy, vz))``.

    Attributes:
        vx: X component.
        vy: Y component.
        vz: Z component.
    """

    def __init__(self, *args):
        """Initialise a Vector3d.

        Args:
            *args: One of the supported signatures described in the class
                docstring.

        Raises:
            Exception: On unsupported argument types or counts.
        """
        if len(args) == 3:
            for a in args:
                if isinstance(a, float) is False and isinstance(a, int) is False:
                    raise Exception("With len args == 3 only float or int type !!!")

            self.vx = float(args[0])
            self.vy = float(args[1])
            self.vz = float(args[2])

        elif len(args) == 2:
            for a in args:
                if not isinstance(a, Point3d):
                    raise Exception("With len args == 2 only Point3d type !!!")

            self.vx = args[1].x - args[0].x
            self.vy = args[1].y - args[0].y
            self.vz = args[1].z - args[0].z

        elif len(args) == 1:
            if not isinstance(args[0], tuple):
                raise Exception("With 1 arg only a tuple !!!")

            if len(args[0]) != 3:
                raise Exception("Tuple must have len == 3 !!!", len(args[0]))

            arg0 = not(not isinstance(args[0][0], float) and not isinstance(args[0][0], int))
            arg1 = not(not isinstance(args[0][1], float) and not isinstance(args[0][1], int))
            arg2 = not(not isinstance(args[0][2], float) and not isinstance(args[0][2], int))

            if not arg0 * arg1 * arg2:
                raise Exception("Tuple must be (float | int, float | int, float | int)!!!")

            self.vx = args[0][0]
            self.vy = args[0][1]
            self.vz = args[0][2]

        elif len(args) == 0:
            self.vx = 0.0
            self.vy = 0.0
            self.vz = 0.0

        else:
            raise Exception("Wrong args !!!", len(args))

    def __str__(self):
        return f"Vector3D: vx = {self.vx:.3e}, vy = {self.vy:.3e}, vz = {self.vz:.3e}"

    def __add__(self, other: Vector3d) -> Vector3d:
        return Vector3d(self.vx + other.vx, self.vy + other.vy, self.vz + other.vz)

    def __sub__(self, other: Vector3d) -> Vector3d:
        return Vector3d(self.vx - other.vx, self.vy - other.vy, self.vz - other.vz)

    # def __mul__(self, scale):
    #    return Vector3d(self.vx*scale,self.vy*scale,self.z*scale)

    def __rmul__(self, scale):
        return Vector3d(self.vx * scale, self.vy * scale, self.vz * scale)

    #       _    _    _            _         _         _
    #     | x    y    z  |   y1*z2*x + z1*x2*y + x1*y2*z +
    #  det| x1   y1   z1 | =       _         _         _
    #     | x2   y2   z2 |  -y2*z1*x - x1*z2*y - x2*y1*z
    #
    def cross(self, other) -> Vector3d:
        """Return the 3-D cross product ``self × other``.

        Args:
            other: The second ``Vector3d``.

        Returns:
            Vector3d: A new vector orthogonal to both operands.
        """
        x1 = self.vx
        y1 = self.vy
        z1 = self.vz
        x2 = other.vx
        y2 = other.vy
        z2 = other.vz
        x3 = y1 * z2 - y2 * z1
        y3 = z1 * x2 - x1 * z2
        z3 = x1 * y2 - x2 * y1
        return Vector3d(x3, y3, z3)

    def norm(self) -> float:
        """Return the Euclidean norm of the vector.

        Returns:
            float: ``sqrt(vx² + vy² + vz²)``.
        """
        return math.sqrt(
            math.pow(self.vx, 2) + math.pow(self.vy, 2) + math.pow(self.vz, 2)
        )

    def normalize(self) -> Vector3d:
        """Normalise the vector in-place to unit length.

        Returns:
            Vector3d: *self* after normalisation (allows chaining).

        Raises:
            Exception: If the vector is the zero vector.
        """
        n = self.norm()
        if n != 0:
            self.vx = self.vx / n
            self.vy = self.vy / n
            self.vz = self.vz / n
        else:
            raise Exception("Vector 3D null !!!")
        return self

    def scalar(self, other) -> float:
        """Return the dot product of this vector with *other*.

        Args:
            other: The second ``Vector3d``.

        Returns:
            float: ``vx*other.vx + vy*other.vy + vz*other.vz``.
        """
        return self.vx * other.vx + self.vy * other.vy + self.vz * other.vz

    def __eq__(self, other: Vector3d) -> bool:
        if not isinstance(other, Vector3d):
            return NotImplemented

        if self.vx == other.vx and self.vy == other.vy and self.vz == other.vz:
            return True

        return False

# Sum vector to point in affine space
def affineSum2d(p: Point2d, v: Vector2d) -> Point2d:
    """Return the point obtained by adding a 2-D vector to a point.

    Args:
        p: Base point.
        v: Displacement vector.

    Returns:
        Point2d: ``p + v``.
    """
    return Point2d(p.x + v.vx, p.y + v.vy)


# Make a points array with equal space between theme
# Extremes stand for include first and last point
def twoPointsDivide(
    p0: Point2d, p1: Point2d, nb: int, extremes: bool = True
) -> List[Point2d]:
    """Divide the segment ``p0 → p1`` into *nb* equal parts.

    Args:
        p0: Start point.
        p1: End point.
        nb: Number of equal sub-divisions (must be ≥ 1).
        extremes: When ``True`` (default), *p0* and *p1* are included as the
            first and last points of the result.

    Returns:
        List[Point2d]: Points along the segment at equal spacing.

    Raises:
        ValueError: If *nb* < 1.
    """
    if nb < 1:
        raise ValueError("nb must be >=1")
    lPoints = []
    if extremes:
        lPoints.append(deepcopy(p0))
    vDir = Vector2d(p0, p1).normalize()
    space = p0.distance(p1) / nb
    for i in range(1, nb):
        lPoints.append(affineSum2d(p0, space * i * vDir))
    if extremes:
        lPoints.append(deepcopy(p1))
    return lPoints


def twoPointsMiddle(p0: Point2d, p1: Point2d) -> Point2d:
    """Return the midpoint between two 2-D points.

    Args:
        p0: First point.
        p1: Second point.

    Returns:
        Point2d: The midpoint ``(p0 + p1) / 2``.
    """
    return Point2d((p0.x + p1.x) / 2, (p0.y + p1.y) / 2)


# Copy of points translated with offset
def twoPointsOffset(
    p0: Point2d, p1: Point2d, offset: float = 0.0
) -> Tuple[Point2d, Point2d]:
    """Return copies of *p0* and *p1* offset perpendicularly by *offset*.

    The normal direction is obtained by rotating the segment direction 90°
    counter-clockwise.

    Args:
        p0: Start point of the reference segment.
        p1: End point of the reference segment.
        offset: Lateral offset distance (positive = left of ``p0→p1``).

    Returns:
        Tuple[Point2d, Point2d]: The offset copies ``(p0', p1')``.
    """
    nDir = offset * Vector2d(p0, p1).normalize().rotate(90)
    return affineSum2d(p0, nDir), affineSum2d(p1, nDir)


class Point2dMdl(BaseModel):
    """Pydantic model wrapping an optional 2-D point coordinate pair.

    Attributes:
        xy: Optional ``(x, y)`` coordinate tuple.
    """

    xy: Optional[Tuple[float, float]] = None


class Vector3dApp(Vector3d):
    """A 3-D vector with an explicit application point (origin).

    Extends ``Vector3d`` by storing the point in space where the vector is
    applied, which is required for moment transport and force-couple
    operations.

    Attributes:
        o: Application point (``Point3d``).
    """

    def __init__(self, origin: Point3d, *args: float | int | tuple[float|int, float|int, float|int]):
        """Initialise a Vector3dApp.

        Args:
            origin: The application point of the vector.
            *args: Vector components forwarded to the ``Vector3d`` constructor
                (three scalars, a 3-tuple, or nothing for the zero vector).
        """
        super().__init__(*args)
        self.__origin = origin

    @property
    def o(self) -> Point3d:
        """Point3d: Application point of the vector."""
        return deepcopy(self.__origin)

    @o.setter
    def o(self, val: Point3d) -> None:
        self.__origin = deepcopy(val)

    def __add__(self, other: Vector3d) -> "Vector3dApp":
        return Vector3dApp(self.__origin, self.vx + other.vx, self.vy + other.vy, self.vz + other.vz)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Vector3dApp):
            return NotImplemented

        c1 = self.vx == other.vx
        c2 = self.vy == other.vy
        c3 = self.vz == other.vz
        c4 = self.__origin.x == other.o.x
        c5 = self.__origin.y == other.o.y
        c6 = self.__origin.z == other.o.z

        if all([c1, c2, c3, c4, c5, c6]):
            return True
        else:
            return False

class ForceCouple:
    """A force-couple system (wrench) consisting of a resultant force and a moment.

    Stores a force applied at a specific point together with a free couple
    vector.  The couple is automatically updated when the application point is
    changed via the :attr:`o` setter (moment transport formula).

    Attributes:
        force: Applied force as a ``Vector3dApp``.
        couple: Free moment vector as a ``Vector3d``.
    """

    def __init__(self, force: Vector3dApp, couple: Vector3d):
        """Initialise a ForceCouple.

        Deep copies are taken of both arguments to prevent external mutation.

        Args:
            force: The applied force (carries its application point).
            couple: The free moment vector.
        """
        self.__force: Vector3dApp = deepcopy(force)
        self.__couple: Vector3d = deepcopy(couple)

    @property
    def force(self) -> Vector3dApp:
        """Vector3dApp: Deep copy of the applied force."""
        return deepcopy(self.__force)

    @force.setter
    def force(self, val: Vector3dApp) -> None:
        self.__force = val

    @property
    def couple(self) -> Vector3d:
        """Vector3d: Deep copy of the free moment vector."""
        return deepcopy(self.__couple)

    @couple.setter
    def couple(self, val: Vector3d) -> None:
        self.__couple = val

    @property
    def o(self) -> Point3d:
        """Point3d: Application point of the force (read from the force vector)."""
        return self.__force.o

    @o.setter
    def o(self, new_o: Point3d) -> None:
        """Transport the force-couple to a new application point.

        Updates the couple via the moment transport formula:
        ``M' = r × F + M``
        where *r* is the vector from *new_o* to the current application point.

        Args:
            new_o: New application point.

        Raises:
            ValueError: If *new_o* is not a ``Point3d``.
        """
        if not isinstance(new_o, Point3d):
            raise ValueError("new_o must be a Point3d !!!", type(new_o))

        old_o = self.__force.o
        v_new_old = Vector3d(
            old_o.x - new_o.x,
            old_o.y - new_o.y,
            old_o.z - new_o.z,
        )
        self.__couple = v_new_old.cross(self.__force) + self.__couple
        self.__force.o = new_o

    def __add__(self, other: ForceCouple) -> ForceCouple:
        """Return the sum of two force-couples reduced to the same point.

        The *other* force-couple is transported to *self*'s application point
        before summing.

        Args:
            other: The second ``ForceCouple``.

        Returns:
            ForceCouple: Resultant force-couple at *self*'s application point.

        Raises:
            ValueError: If *other* is not a ``ForceCouple``.
        """
        if not isinstance(other, ForceCouple):
            raise ValueError("new_o must be a ForceCouple !!!", type(other))
        self.__force += other.force
        other.o = self.o
        self.__couple += other.couple
        return ForceCouple(self.__force, self.__couple)


class ForceCoupleSystem:
    """A system of multiple ``ForceCouple`` objects.

    Provides utilities to compute the barycentre of application points and to
    reduce the system to a single equivalent force-couple at a given point.

    Note:
        This class is currently a stub — methods are not yet implemented.
    """

    def __init__(self, forces: List[ForceCouple]):
        """Initialise a ForceCoupleSystem.

        Args:
            forces: List of ``ForceCouple`` objects forming the system.
        """
        self.__fc = forces

    def baryOfApplicationPoints(self) -> Point3d:
        """Return the barycentre of all force application points.

        Returns:
            Point3d: The barycentre point.
        """
        nb_points = len(self.__fc)
        xg = sum([f.o.x for f in self.__fc]) / nb_points
        yg = sum([f.o.y for f in self.__fc]) / nb_points
        zg = sum([f.o.z for f in self.__fc]) / nb_points

        return Point3d(xg, yg, zg)


    def forceCoupleResume(self, applicationPoint: Point3d) -> ForceCouple:
        """Reduce the entire system to a single force-couple at *applicationPoint*.

        Args:
            applicationPoint: The reference point for moment transport.

        Returns:
            ForceCouple: Equivalent resultant force-couple.
        """
        forces = deepcopy(self.__fc)
        resumed = forces[0]
        resumed.o = applicationPoint
        for i in range(1, len(forces)):
            resumed += forces[i]

        return resumed


# import math
from pathlib import Path

# import pytest
from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.materials import Concrete, ConcreteSteel
from pycivil.EXAStructural.templateRCRect import RCTemplRectEC2, SectionPlot

from matplotlib import (
    pyplot as plt,
)

from matplotlib.pyplot import Figure, Axes

def section_1():
    code = Code("NTC2008")
    concrete = Concrete(descr="My concrete")
    concrete.setByCode(code, "C32/40")
    steel = ConcreteSteel(descr="My steel")
    steel.setByCode(code, "B450C")
    rcSection = RCTemplRectEC2(1, "Template RC Section - Manuscript #5")
    rcSection.setDimH(500)
    rcSection.setDimW(1000)
    rcSection.addSteelArea("LINE-MT", dist=74, d=24, nb=5, sd=200)
    rcSection.addSteelArea("LINE-MB", dist=71, d=18, nb=5, sd=200)
    return rcSection

def test_section_1(tmp_path: Path):
    section_plotter = SectionPlot(section_1())
    section_plotter.plot()
    section_plotter.save(tmp_path / Path("test.png"), dpi=300)



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

from pycivil.EXAStructural.codes import Code
from pycivil.EXAUtils.typstReportMakers import CodesTypstFB
from pycivil.EXAUtils.report import (
    TypstReporter,
    ReportTemplateEnum,
    getTypstTemplatesPath,
)


def test_report_typst_codes(tmp_path: Path):
    """Test codes/bibliography typst fragment generation."""
    c1 = Code("NTC2008")
    c2 = Code("EC2:ITA")
    c3 = Code("CIRC2008")
    c4 = Code("CIRC2018")
    c5 = Code("EC2")

    typstTemplatePath = getTypstTemplatesPath()
    codes = CodesTypstFB(typstTemplatePath, [c1, c2, c3])
    codes.appendUniqueCode(c4)
    codes.appendUniqueCode(c3)
    codes.appendUniqueCode(c5)

    codes.setOptionNameList("Laws used")
    codes.setOptionSection()

    reporter = TypstReporter(typstTemplatePath)
    reporter.linkFragments(
        template=ReportTemplateEnum.TYP_ENG_CAL,
        fragments=[codes.buildFragment()],
    )
    reporter.compileDocument(path=str(tmp_path), output_format="pdf")
    assert tmp_path.joinpath("report.typ").exists()
    assert tmp_path.joinpath("report.pdf").exists()
    assert tmp_path.joinpath("report.pdf").stat().st_size > 0

    # Verify content has code references
    content = tmp_path.joinpath("report.typ").read_text(encoding="utf-8")
    assert "Laws used" in content



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

"""Markdown report generation tests for RC section checks.

These tests mirror test_report_latex_rcsec_check.py but generate
markdown and HTML output instead of LaTeX/PDF.
"""

import json
from pathlib import Path

import pytest

from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import (
    ModelInputRcSecCheck,
    ModelOutputRcSecCheck,
    RcSecRectSolver,
)
from pycivil.EXAUtils.report import (
    MarkdownReporter,
    ReportProperties,
    ReportTemplateEnum,
    getMarkdownTemplatesPath,
)


def test_report_markdown_rcsec_check_01(tmp_path: Path):
    """Test markdown report generation for RC section check case 01."""
    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_in_01.json"
    ) as jsonFile:
        jsonObjectIn = json.load(jsonFile)

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_out_01.json"
    ) as jsonFile:
        jsonObjectOut = json.load(jsonFile)

    iData = ModelInputRcSecCheck(**jsonObjectIn)
    oData = ModelOutputRcSecCheck(**jsonObjectOut)

    markdownTemplatePath = getMarkdownTemplatesPath()
    fragments = RcSecRectSolver.buildMarkdownReport(iData, oData, markdownTemplatePath)

    reporter = MarkdownReporter(markdownTemplatePath)
    reporter.setProperties(ReportProperties(
        project_brief="RC Section Check - Test 01",
        module_name="PyCivil",
        module_version="0.2.16",
        report_designer="Test Engineer",
        report_date="2026-02-03",
        report_time="10:00",
        report_token="MD-RCSEC-01",
    ))
    reporter.linkFragments(template=ReportTemplateEnum.MD_ENG_CAL, fragments=fragments)
    reporter.compileDocument(path=str(tmp_path), fileName="report", output_format="md")

    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.md").exists()

    # Verify content
    content = tmp_path.joinpath("report.md").read_text(encoding="utf-8")
    assert "Geometria della sezione" in content
    assert "Calcestruzzo" in content
    assert "Acciaio" in content


def test_report_markdown_rcsec_check_01_html(tmp_path: Path):
    """Test HTML report generation for RC section check case 01."""
    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_in_01.json"
    ) as jsonFile:
        jsonObjectIn = json.load(jsonFile)

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_out_01.json"
    ) as jsonFile:
        jsonObjectOut = json.load(jsonFile)

    iData = ModelInputRcSecCheck(**jsonObjectIn)
    oData = ModelOutputRcSecCheck(**jsonObjectOut)

    markdownTemplatePath = getMarkdownTemplatesPath()
    fragments = RcSecRectSolver.buildMarkdownReport(iData, oData, markdownTemplatePath)

    reporter = MarkdownReporter(markdownTemplatePath)
    reporter.setProperties(ReportProperties(
        project_brief="RC Section Check - Test 01 HTML",
        module_name="PyCivil",
        module_version="0.2.16",
        report_designer="Test Engineer",
        report_date="2026-02-03",
        report_time="10:30",
        report_token="HTML-RCSEC-01",
    ))
    reporter.linkFragments(template=ReportTemplateEnum.MD_ENG_CAL, fragments=fragments)
    reporter.compileDocument(path=str(tmp_path), fileName="report", output_format="html")

    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.md").exists()
    assert tmp_path.joinpath("report.html").exists()

    # Verify HTML content
    html_content = tmp_path.joinpath("report.html").read_text(encoding="utf-8")
    assert "<html>" in html_content
    assert "MathJax" in html_content  # Check that MathJax is included for formulas
    assert "<table>" in html_content or "<th>" in html_content


def test_report_markdown_rcsec_check_02(tmp_path: Path):
    """Test markdown report generation for RC section check case 02."""
    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_in_02.json"
    ) as jsonFile:
        jsonObjectIn = json.load(jsonFile)

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_out_02.json"
    ) as jsonFile:
        jsonObjectOut = json.load(jsonFile)

    iData = ModelInputRcSecCheck(**jsonObjectIn)
    oData = ModelOutputRcSecCheck(**jsonObjectOut)

    markdownTemplatePath = getMarkdownTemplatesPath()
    fragments = RcSecRectSolver.buildMarkdownReport(iData, oData, markdownTemplatePath)

    reporter = MarkdownReporter(markdownTemplatePath)
    reporter.setProperties(ReportProperties(
        project_brief="RC Section Check - Test 02",
        module_name="PyCivil",
        module_version="0.2.16",
        report_designer="Test Engineer",
        report_date="2026-02-03",
        report_time="11:00",
        report_token="MD-RCSEC-02",
    ))
    reporter.linkFragments(template=ReportTemplateEnum.MD_ENG_CAL, fragments=fragments)
    reporter.compileDocument(path=str(tmp_path), fileName="report", output_format="md")

    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.md").exists()


def test_report_markdown_rcsec_check_02_html(tmp_path: Path):
    """Test HTML report generation for RC section check case 02."""
    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_in_02.json"
    ) as jsonFile:
        jsonObjectIn = json.load(jsonFile)

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_out_02.json"
    ) as jsonFile:
        jsonObjectOut = json.load(jsonFile)

    iData = ModelInputRcSecCheck(**jsonObjectIn)
    oData = ModelOutputRcSecCheck(**jsonObjectOut)

    markdownTemplatePath = getMarkdownTemplatesPath()
    fragments = RcSecRectSolver.buildMarkdownReport(iData, oData, markdownTemplatePath)

    reporter = MarkdownReporter(markdownTemplatePath)
    reporter.setProperties(ReportProperties(
        project_brief="RC Section Check - Test 02 HTML",
        module_name="PyCivil",
        module_version="0.2.16",
        report_designer="Test Engineer",
        report_date="2026-02-03",
        report_time="11:30",
        report_token="HTML-RCSEC-02",
    ))
    reporter.linkFragments(template=ReportTemplateEnum.MD_ENG_CAL, fragments=fragments)
    reporter.compileDocument(path=str(tmp_path), fileName="report", output_format="html")

    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.md").exists()
    assert tmp_path.joinpath("report.html").exists()



# Copyright (c) 2026 Luigi Paone <ppc.luigi.paone@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

"""Typst report generation tests for RC section checks.

These tests mirror test_report_markdown_rcsec_check.py but generate
Typst and PDF output instead of Markdown/HTML.
"""

import json
from pathlib import Path

from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import (
    ModelInputRcSecCheck,
    ModelOutputRcSecCheck,
    RcSecRectSolver,
)
from pycivil.EXAStructural.templateRCRect import SectionPlot
from pycivil.EXAUtils.report import (
    TypstReporter,
    ReportProperties,
    ReportTemplateEnum,
    getTypstTemplatesPath,
)


def _create_section_png(iData: ModelInputRcSecCheck, path: Path) -> None:
    """Create section plot PNG for Typst templates."""
    section = iData.buildRcsRectangular()
    if section is not None:
        plotter = SectionPlot(section)
        plotter.plot()
        plotter.save(path / "section.png")


def test_report_typst_rcsec_check_01(tmp_path: Path):
    """Test typst report generation for RC section check case 01."""
    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_in_01.json"
    ) as jsonFile:
        jsonObjectIn = json.load(jsonFile)

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_out_01.json"
    ) as jsonFile:
        jsonObjectOut = json.load(jsonFile)

    iData = ModelInputRcSecCheck(**jsonObjectIn)
    oData = ModelOutputRcSecCheck(**jsonObjectOut)

    # Create section PNG
    _create_section_png(iData, tmp_path)

    typstTemplatePath = getTypstTemplatesPath()
    fragments = RcSecRectSolver.buildTypstReport(iData, oData, typstTemplatePath)

    reporter = TypstReporter(typstTemplatePath)
    reporter.setProperties(ReportProperties(
        project_brief="RC Section Check - Test 01",
        module_name="PyCivil",
        module_version="0.2.20",
        report_designer="Test Engineer",
        report_date="2026-02-18",
        report_time="10:00",
        report_token="TYP-RCSEC-01",
    ))
    reporter.linkFragments(template=ReportTemplateEnum.TYP_ENG_CAL, fragments=fragments)
    reporter.compileDocument(path=str(tmp_path), fileName="report", output_format="pdf")

    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.typ").exists()
    assert tmp_path.joinpath("report.pdf").exists()
    assert tmp_path.joinpath("report.pdf").stat().st_size > 0

    # Verify content
    content = tmp_path.joinpath("report.typ").read_text(encoding="utf-8")
    assert "Geometria della sezione" in content
    assert "Calcestruzzo" in content
    assert "Acciaio" in content


def test_report_typst_rcsec_check_02(tmp_path: Path):
    """Test typst report generation for RC section check case 02."""
    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_in_02.json"
    ) as jsonFile:
        jsonObjectIn = json.load(jsonFile)

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_out_02.json"
    ) as jsonFile:
        jsonObjectOut = json.load(jsonFile)

    iData = ModelInputRcSecCheck(**jsonObjectIn)
    oData = ModelOutputRcSecCheck(**jsonObjectOut)

    # Create section PNG
    _create_section_png(iData, tmp_path)

    typstTemplatePath = getTypstTemplatesPath()
    fragments = RcSecRectSolver.buildTypstReport(iData, oData, typstTemplatePath)

    reporter = TypstReporter(typstTemplatePath)
    reporter.setProperties(ReportProperties(
        project_brief="RC Section Check - Test 02",
        module_name="PyCivil",
        module_version="0.2.20",
        report_designer="Test Engineer",
        report_date="2026-02-18",
        report_time="11:00",
        report_token="TYP-RCSEC-02",
    ))
    reporter.linkFragments(template=ReportTemplateEnum.TYP_ENG_CAL, fragments=fragments)
    reporter.compileDocument(path=str(tmp_path), fileName="report", output_format="pdf")

    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.typ").exists()
    assert tmp_path.joinpath("report.pdf").exists()
    assert tmp_path.joinpath("report.pdf").stat().st_size > 0
